import time
import warnings
from typing import Any, Dict, List, Optional, Union

from flet.core.alignment import Alignment
from flet.core.animation import AnimationValue
from flet.core.buttons import ButtonStyle
from flet.core.control import Control, OptionalNumber
from flet.core.form_field_control import FormFieldControl, InputBorder
from flet.core.menu_bar import MenuStyle
from flet.core.ref import Ref
from flet.core.text_style import TextStyle
from flet.core.textfield import InputFilter, TextCapitalization
from flet.core.types import (
    BorderRadiusValue,
    ColorValue,
    ControlState,
    ControlStateValue,
    IconEnums,
    IconValueOrControl,
    Number,
    OffsetValue,
    OptionalControlEventCallable,
    OptionalEventCallable,
    PaddingValue,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
    TextAlign,
)


class Option(Control):
    def __init__(
        self,
        key: Optional[str] = None,
        text: Optional[str] = None,
        content: Optional[Control] = None,
        leading_icon: Optional[IconValueOrControl] = None,
        trailing_icon: Optional[IconValueOrControl] = None,
        style: Optional[ButtonStyle] = None,
        alignment: Optional[Alignment] = None,  # to be deprecated
        text_style: Optional[TextStyle] = None,  # to be deprecated
        on_click: OptionalControlEventCallable = None,  # to be deprecated
        #
        # Control
        #
        ref=None,
        disabled: Optional[bool] = None,
        visible: Optional[bool] = None,
        data: Any = None,
    ):
        Control.__init__(self, ref=ref, disabled=disabled, visible=visible, data=data)
        self.key = key
        self.text = text
        self.content = content
        self.leading_icon = leading_icon
        self.trailing_icon = trailing_icon
        self.style = style

        deprecated_properties_list = ["text_style", "on_click", "alignment"]

        for item in deprecated_properties_list:
            if eval(item) is not None:
                warnings.warn(
                    f"{item} is deprecated since version 0.27.0 "
                    f"and will be removed in version 0.30.0.",
                    category=DeprecationWarning,
                    stacklevel=2,
                )

    def _get_control_name(self):
        return "dropdownoption"

    def _get_children(self):
        children = super()._get_children()
        if isinstance(self.__leading_icon, Control):
            self.__leading_icon._set_attr_internal("n", "leadingIcon")
            children.append(self.__leading_icon)
        if isinstance(self.__trailing_icon, Control):
            self.__trailing_icon._set_attr_internal("n", "trailing_icon")
            children.append(self.__trailing_icon)
        if isinstance(self.__content, Control):
            self.__content._set_attr_internal("n", "content")
            children.append(self.__content)
        return children

    def before_update(self):
        super().before_update()
        assert (
            self.key is not None or self.text is not None
        ), "key or text must be specified"
        self._set_attr_json("style", self.__style)

    # key
    @property
    def key(self) -> Optional[str]:
        return self._get_attr("key")

    @key.setter
    def key(self, value: Optional[str]):
        self._set_attr("key", value)

    # text
    @property
    def text(self) -> Optional[str]:
        return self._get_attr("text")

    @text.setter
    def text(self, value: Optional[str]):
        self._set_attr("text", value)

    # content
    @property
    def content(self) -> Optional[Control]:
        return self.__content

    @content.setter
    def content(self, value: Optional[Control]):
        self.__content = value

    # style
    @property
    def style(self) -> Optional[ButtonStyle]:
        return self.__style

    @style.setter
    def style(self, value: Optional[ButtonStyle]):
        self.__style = value

    # leading_icon
    @property
    def leading_icon(self) -> Optional[IconValueOrControl]:
        return self.__leading_icon

    @leading_icon.setter
    def leading_icon(self, value: Optional[IconValueOrControl]):
        self.__leading_icon = value
        if not isinstance(value, Control):
            self._set_enum_attr("leadingIcon", value, IconEnums)

    # trailing_icon
    @property
    def trailing_icon(self) -> Optional[IconValueOrControl]:
        return self.__trailing_icon

    @trailing_icon.setter
    def trailing_icon(self, value: Optional[IconValueOrControl]):
        self.__trailing_icon = value
        if not isinstance(value, Control):
            self._set_enum_attr("trailingIcon", value, IconEnums)


class DropdownOption(Option):
    """Alias for Option"""


class Dropdown(FormFieldControl):
    """
    A dropdown control that allows users to select a single option from a list of options.
    -----
    Online docs: https://flet.dev/docs/controls/dropdown
    """

    def __init__(
        self,
        value: Optional[str] = None,
        autofocus: Optional[bool] = None,
        text_align: Optional[TextAlign] = None,
        elevation: ControlStateValue[OptionalNumber] = None,
        options: Optional[List[Option]] = None,
        label_content: Optional[str] = None,
        enable_filter: Optional[bool] = None,
        enable_search: Optional[bool] = None,
        editable: Optional[bool] = None,
        max_menu_height: OptionalNumber = None,  # to be discontinued
        menu_height: OptionalNumber = None,
        menu_width: OptionalNumber = None,
        expanded_insets: PaddingValue = None,
        selected_suffix: Optional[Control] = None,
        input_filter: Optional[InputFilter] = None,
        capitalization: Optional[TextCapitalization] = None,
        options_fill_horizontally: Optional[bool] = None,  # to be deprecated
        padding: Optional[PaddingValue] = None,  # to be deprecated
        trailing_icon: Optional[IconValueOrControl] = None,
        leading_icon: Optional[IconValueOrControl] = None,
        select_icon: Optional[IconValueOrControl] = None,  # to be deprecated
        selected_trailing_icon: Optional[IconValueOrControl] = None,
        on_change: OptionalEventCallable = None,
        on_focus: OptionalEventCallable = None,
        on_blur: OptionalEventCallable = None,
        enable_feedback: Optional[bool] = None,  # to be deprecated
        item_height: OptionalNumber = None,  # to be deprecated
        alignment: Optional[Alignment] = None,  # to be deprecated
        hint_content: Optional[Control] = None,  # to be deprecated
        icon_content: Optional[Control] = None,  # to be deprecated
        select_icon_size: OptionalNumber = None,  # to be deprecated
        icon_size: OptionalNumber = None,  # to be deprecated
        select_icon_enabled_color: Optional[ColorValue] = None,  # to be deprecated
        icon_enabled_color: Optional[ColorValue] = None,  # to be deprecated
        select_icon_disabled_color: Optional[ColorValue] = None,  # to be deprecated
        icon_disabled_color: Optional[ColorValue] = None,  # to be deprecated
        #
        # FormField specific
        #
        bgcolor: Optional[ColorValue] = None,
        error_style: Optional[TextStyle] = None,
        error_text: Optional[str] = None,
        text_size: OptionalNumber = None,
        text_style: Optional[TextStyle] = None,
        label: Optional[str] = None,
        label_style: Optional[TextStyle] = None,
        icon: Optional[IconValueOrControl] = None,  # to deprecated
        border: Optional[InputBorder] = None,
        color: Optional[str] = None,
        focused_color: Optional[str] = None,  # to be deprecated
        focused_bgcolor: Optional[str] = None,  # to be deprecated
        border_width: OptionalNumber = None,
        border_color: Optional[str] = None,
        border_radius: Optional[BorderRadiusValue] = None,
        focused_border_width: OptionalNumber = None,
        focused_border_color: Optional[str] = None,
        content_padding: PaddingValue = None,
        dense: Optional[bool] = None,
        filled: Optional[bool] = None,
        fill_color: Optional[str] = None,
        hover_color: Optional[str] = None,
        hint_text: Optional[str] = None,
        hint_style: Optional[TextStyle] = None,
        helper_text: Optional[str] = None,
        helper_style: Optional[TextStyle] = None,
        prefix: Optional[Control] = None,  # to be deprecated
        prefix_text: Optional[str] = None,  # to be deprecated
        prefix_style: Optional[TextStyle] = None,  # to be deprecated
        prefix_icon: Optional[str] = None,  # to be deprecated
        disabled_hint_content: Optional[Control] = None,  # to be deprecated
        suffix: Optional[Control] = None,  # to be deprecated
        suffix_icon: Optional[IconValueOrControl] = None,  # to be deprecated
        suffix_text: Optional[str] = None,  # to be deprecated
        suffix_style: Optional[TextStyle] = None,  # to be deprecated
        counter: Optional[Control] = None,  # to be deprecated
        counter_text: Optional[str] = None,  # to be deprecated
        counter_style: Optional[TextStyle] = None,  # to be deprecated
        #
        # ConstrainedControl
        #
        ref: Optional[Ref] = None,
        key: Optional[str] = None,
        width: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: RotateValue = None,
        scale: ScaleValue = None,
        offset: OffsetValue = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: AnimationValue = None,
        animate_size: AnimationValue = None,
        animate_position: AnimationValue = None,
        animate_rotation: AnimationValue = None,
        animate_scale: AnimationValue = None,
        animate_offset: AnimationValue = None,
        on_animation_end: OptionalEventCallable = None,
        tooltip: Optional[str] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        FormFieldControl.__init__(
            self,
            text_size=text_size,
            text_style=text_style,
            label=label,
            label_style=label_style,
            border=border,
            color=color,
            border_width=border_width,
            border_color=border_color,
            border_radius=border_radius,
            focused_border_width=focused_border_width,
            focused_border_color=focused_border_color,
            content_padding=content_padding,
            dense=dense,
            filled=filled,
            fill_color=fill_color,
            hover_color=hover_color,
            hint_text=hint_text,
            hint_style=hint_style,
            helper_text=helper_text,
            helper_style=helper_style,
            error_text=error_text,
            error_style=error_style,
            prefix_icon=prefix_icon,
            ref=ref,
            key=key,
            width=width,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            tooltip=tooltip,
            visible=visible,
            disabled=disabled,
            data=data,
        )

        deprecated_properties_list = [
            "select_icon_size",
            "select_icon_enabled_color",
            "select_icon_disabled_color",
            "suffix",
            "suffix_icon",
            "suffix_style",
            "suffix_text",
            "icon_content",
            "icon_enabled_color",
            "icon_disabled_color",
            "icon_size",
            "icon",
            "hint_content",
            "prefix_text",
            "prefix_style",
            "prefix",
            "prefix_icon",
            "focused_color",
            "disabled_hint_content",
            "alignment",
            "focused_bgcolor",
            "item_height",
            "enable_feedback",
            "options_fill_horizontally",
            "padding",
            "max_menu_height",
        ]

        for item in deprecated_properties_list:
            if eval(item) is not None:
                warnings.warn(
                    f"{item} is deprecated since version 0.27.0 "
                    f"and will be removed in version 0.30.0.",
                    category=DeprecationWarning,
                    stacklevel=2,
                )

        self.options = options
        self.selected_suffix = selected_suffix
        self.input_filter = input_filter
        self.enable_filter = enable_filter
        self.enable_search = enable_search
        self.editable = editable
        self.menu_height = menu_height
        self.menu_width = menu_width
        self.expanded_insets = expanded_insets
        self.capitalization = capitalization
        self.label_content = label_content
        self.leading_icon = leading_icon
        self.trailing_icon = trailing_icon
        self.selected_trailing_icon = selected_trailing_icon
        self.select_icon = select_icon
        self.on_change = on_change
        self.on_focus = on_focus
        self.on_blur = on_blur
        self.value = value
        self.bgcolor = bgcolor
        self.elevation = elevation
        self.text_align = text_align
        self.autofocus = autofocus

    def _get_control_name(self):
        return "dropdown"

    def before_update(self):
        super().before_update()
        self._set_attr_json("bgcolor", self.__bgcolor, wrap_attr_dict=True)
        self._set_attr_json("elevation", self.__elevation, wrap_attr_dict=True)
        ##self._set_attr_json("inputFilter", self.__input_filter)
        ##self._set_attr_json("expandInsets", self.__expanded_insets)
        self.expand_loose = self.expand  # to fix a display issue

    def _get_children(self):
        children = FormFieldControl._get_children(self) + self.__options
        if isinstance(self.__leading_icon, Control):
            self.__leading_icon._set_attr_internal("n", "leading_icon")
            children.append(self.__leading_icon)
        if isinstance(self.__select_icon, Control):
            self.__select_icon._set_attr_internal("n", "select_icon")
            children.append(self.__select_icon)
        if isinstance(self.__trailing_icon, Control):
            self.__trailing_icon._set_attr_internal("n", "trailing_icon")
            children.append(self.__trailing_icon)
        if isinstance(self.__selected_trailing_icon, Control):
            self.__selected_trailing_icon._set_attr_internal(
                "n", "selected_trailing_icon"
            )
            children.append(self.__selected_trailing_icon)
        return children

    def __contains__(self, item):
        return item in self.__options

    def focus(self):
        self._set_attr_json("focus", str(time.time()))
        self.update()

    # value
    @property
    def value(self) -> Optional[str]:
        return self._get_attr("value")

    @value.setter
    def value(self, value: Optional[str]):
        self._set_attr("value", value)

    # options
    @property
    def options(self) -> Optional[List[Option]]:
        return self.__options

    @options.setter
    def options(self, value: Optional[List[Option]]):
        self.__options = value if value is not None else []

    # menu_height
    @property
    def menu_height(self) -> OptionalNumber:
        return self._get_attr("menuHeight", data_type="float")

    @menu_height.setter
    def menu_height(self, value: OptionalNumber):
        self._set_attr("menuHeight", value)

    # menu_width
    @property
    def menu_width(self) -> OptionalNumber:
        return self._get_attr("menuWidth", data_type="float")

    @menu_width.setter
    def menu_width(self, value: OptionalNumber):
        self._set_attr("menuWidth", value)

    # editable
    @property
    def editable(self) -> bool:
        return self._get_attr("editable", data_type="bool", def_value=False)

    @editable.setter
    def editable(self, value: Optional[bool]):
        self._set_attr("editable", value)

    # select_icon
    @property
    def select_icon(self) -> Optional[IconValueOrControl]:
        warnings.warn(
            f"select_icon is deprecated since version 0.27.0 "
            f"and will be removed in version 0.30.0. Use trailing_icon instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.__select_icon

    @select_icon.setter
    def select_icon(self, value: Optional[IconValueOrControl]):
        self.__select_icon = value

        if not isinstance(value, Control):
            self._set_enum_attr("selectIcon", value, IconEnums)

        if value is not None:
            warnings.warn(
                f"select_icon is deprecated since version 0.27.0 "
                f"and will be removed in version 0.30.0. Use trailing_icon instead.",
                category=DeprecationWarning,
                stacklevel=2,
            )

    # leading_icon
    @property
    def leading_icon(self) -> Optional[IconValueOrControl]:
        return self.__leading_icon

    @leading_icon.setter
    def leading_icon(self, value: Optional[IconValueOrControl]):
        self.__leading_icon = value
        if not isinstance(value, Control):
            self._set_enum_attr("leadingIcon", value, IconEnums)

    # trailing_icon
    @property
    def trailing_icon(self) -> Optional[IconValueOrControl]:
        return self.__trailing_icon

    @trailing_icon.setter
    def trailing_icon(self, value: Optional[IconValueOrControl]):
        self.__trailing_icon = value
        if not isinstance(value, Control):
            self._set_enum_attr("trailingIcon", value, IconEnums)

    # selected_trailing_icon
    @property
    def selected_trailing_icon(self) -> Optional[IconValueOrControl]:
        return self.__selected_trailing_icon

    @selected_trailing_icon.setter
    def selected_trailing_icon(self, value: Optional[IconValueOrControl]):
        self.__selected_trailing_icon = value
        if not isinstance(value, Control):
            self._set_enum_attr("selectedTrailingIcon", value, IconEnums)

    # bgcolor
    @property
    def bgcolor(self) -> ControlStateValue[ColorValue]:
        return self.__bgcolor

    @bgcolor.setter
    def bgcolor(self, value: ControlStateValue[ColorValue]):
        self.__bgcolor = value

    # text_align
    @property
    def text_align(self) -> Optional[TextAlign]:
        return self.__text_align

    @text_align.setter
    def text_align(self, value: Optional[TextAlign]):
        self.__text_align = value
        self._set_enum_attr("textAlign", value, TextAlign)

    # elevation
    @property
    def elevation(self) -> Union[OptionalNumber, Dict[ControlState, Number]]:
        return self.__elevation

    @elevation.setter
    def elevation(self, value: Union[OptionalNumber, Dict[ControlState, Number]]):
        self.__elevation = value

    # enable_filter
    @property
    def enable_filter(self) -> bool:
        return self._get_attr("enableFilter", data_type="bool", def_value=False)

    @enable_filter.setter
    def enable_filter(self, value: Optional[bool]):
        self._set_attr("enableFilter", value)

    # enable_search
    @property
    def enable_search(self) -> bool:
        return self._get_attr("enableSearch", data_type="bool", def_value=True)

    @enable_search.setter
    def enable_search(self, value: Optional[bool]):
        self._set_attr("enableSearch", value)

    # autofocus
    @property
    def autofocus(self) -> bool:
        return self._get_attr("autofocus", data_type="bool", def_value=False)

    @autofocus.setter
    def autofocus(self, value: Optional[bool]):
        self._set_attr("autofocus", value)

    # on_change
    @property
    def on_change(self) -> OptionalControlEventCallable:
        return self._get_event_handler("change")

    @on_change.setter
    def on_change(self, handler: OptionalControlEventCallable):
        self._add_event_handler("change", handler)

    # on_focus
    @property
    def on_focus(self) -> OptionalControlEventCallable:
        return self._get_event_handler("focus")

    @on_focus.setter
    def on_focus(self, handler: OptionalControlEventCallable):
        self._add_event_handler("focus", handler)

    # on_blur
    @property
    def on_blur(self) -> OptionalControlEventCallable:
        return self._get_event_handler("blur")

    @on_blur.setter
    def on_blur(self, handler: OptionalControlEventCallable):
        self._add_event_handler("blur", handler)



import time
from typing import Any, Optional, Union

from flet.core.adaptive_control import AdaptiveControl
from flet.core.animation import AnimationValue
from flet.core.badge import BadgeValue
from flet.core.buttons import ButtonStyle
from flet.core.constrained_control import ConstrainedControl
from flet.core.control import Control, OptionalNumber
from flet.core.ref import Ref
from flet.core.tooltip import TooltipValue
from flet.core.types import (
    ClipBehavior,
    ColorEnums,
    ColorValue,
    IconEnums,
    IconValue,
    OffsetValue,
    OptionalControlEventCallable,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
    UrlTarget,
)


class OutlinedButton(ConstrainedControl, AdaptiveControl):
    """
    Outlined buttons are medium-emphasis buttons. They contain actions that are important, but aren’t the primary action in an app. Outlined buttons pair well with filled buttons to indicate an alternative, secondary action.

    Example:
    ```
    import flet as ft

    def main(page: ft.Page):
        page.title = "Basic outlined buttons"
        page.add(
            ft.OutlinedButton(text="Outlined button"),
            ft.OutlinedButton("Disabled button", disabled=True),
        )

    ft.app(target=main)
    ```

    -----

    Online docs: https://flet.dev/docs/controls/outlinedbutton
    """

    def __init__(
        self,
        text: Optional[str] = None,
        icon: Optional[IconValue] = None,
        icon_color: Optional[ColorValue] = None,
        content: Optional[Control] = None,
        style: Optional[ButtonStyle] = None,
        autofocus: Optional[bool] = None,
        clip_behavior: Optional[ClipBehavior] = None,
        url: Optional[str] = None,
        url_target: Optional[UrlTarget] = None,
        on_click: OptionalControlEventCallable = None,
        on_long_press: OptionalControlEventCallable = None,
        on_hover: OptionalControlEventCallable = None,
        on_focus: OptionalControlEventCallable = None,
        on_blur: OptionalControlEventCallable = None,
        #
        # ConstrainedControl and AdaptiveControl
        #
        ref: Optional[Ref] = None,
        key: Optional[str] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        left: OptionalNumber = None,
        top: OptionalNumber = None,
        right: OptionalNumber = None,
        bottom: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: OptionalControlEventCallable = None,
        tooltip: Optional[TooltipValue] = None,
        badge: Optional[BadgeValue] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
        adaptive: Optional[bool] = None,
    ):
        ConstrainedControl.__init__(
            self,
            ref=ref,
            key=key,
            width=width,
            height=height,
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            tooltip=tooltip,
            badge=badge,
            visible=visible,
            disabled=disabled,
            data=data,
        )
        AdaptiveControl.__init__(self, adaptive=adaptive)

        self.text = text
        self.icon = icon
        self.icon_color = icon_color
        self.style = style
        self.content = content
        self.autofocus = autofocus
        self.url = url
        self.url_target = url_target
        self.clip_behavior = clip_behavior
        self.on_click = on_click
        self.on_long_press = on_long_press
        self.on_hover = on_hover
        self.on_focus = on_focus
        self.on_blur = on_blur

    def _get_control_name(self):
        return "outlinedbutton"

    def before_update(self):
        super().before_update()
        assert (
            self.text or self.icon or (self.__content and self.__content.visible)
        ), "at minimum, text, icon or a visible content must be provided"
        if self.__style is not None:
            self.__style.side = self._wrap_attr_dict(self.__style.side)
            self.__style.shape = self._wrap_attr_dict(self.__style.shape)
            self.__style.padding = self._wrap_attr_dict(self.__style.padding)
        self._set_attr_json("style", self.__style)

    def _get_children(self):
        if self.__content is None:
            return []
        self.__content._set_attr_internal("n", "content")
        return [self.__content]

    def focus(self):
        self._set_attr_json("focus", str(time.time()))
        self.update()

    # text
    @property
    def text(self) -> Optional[str]:
        return self._get_attr("text")

    @text.setter
    def text(self, value: Optional[str]):
        self._set_attr("text", value)

    # icon
    @property
    def icon(self) -> Optional[IconValue]:
        return self.__icon

    @icon.setter
    def icon(self, value: Optional[IconValue]):
        self.__icon = value
        self._set_enum_attr("icon", value, IconEnums)

    # icon_color
    @property
    def icon_color(self) -> Optional[ColorValue]:
        return self.__icon_color

    @icon_color.setter
    def icon_color(self, value: Optional[ColorValue]):
        self.__icon_color = value
        self._set_enum_attr("iconColor", value, ColorEnums)

    # style
    @property
    def style(self) -> Optional[ButtonStyle]:
        return self.__style

    @style.setter
    def style(self, value: Optional[ButtonStyle]):
        self.__style = value

    # url
    @property
    def url(self) -> Optional[str]:
        return self._get_attr("url")

    @url.setter
    def url(self, value: Optional[str]):
        self._set_attr("url", value)

    # url_target
    @property
    def url_target(self) -> Optional[UrlTarget]:
        return self.__url_target

    @url_target.setter
    def url_target(self, value: Optional[UrlTarget]):
        self.__url_target = value
        self._set_enum_attr("urlTarget", value, UrlTarget)

    # on_click
    @property
    def on_click(self) -> OptionalControlEventCallable:
        return self._get_event_handler("click")

    @on_click.setter
    def on_click(self, handler: OptionalControlEventCallable):
        self._add_event_handler("click", handler)

    # on_long_press
    @property
    def on_long_press(self) -> OptionalControlEventCallable:
        return self._get_event_handler("long_press")

    @on_long_press.setter
    def on_long_press(self, handler: OptionalControlEventCallable):
        self._add_event_handler("long_press", handler)
        self._set_attr("onLongPress", True if handler is not None else None)

    # content
    @property
    def content(self) -> Optional[Control]:
        return self.__content

    @content.setter
    def content(self, value: Optional[Control]):
        self.__content = value

    # autofocus
    @property
    def autofocus(self) -> bool:
        return self._get_attr("autofocus", data_type="bool", def_value=False)

    @autofocus.setter
    def autofocus(self, value: Optional[bool]):
        self._set_attr("autofocus", value)

    # clip_behavior
    @property
    def clip_behavior(self) -> Optional[ClipBehavior]:
        return self.__clip_behavior

    @clip_behavior.setter
    def clip_behavior(self, value: Optional[ClipBehavior]):
        self.__clip_behavior = value
        self._set_enum_attr("clipBehavior", value, ClipBehavior)

    # on_hover
    @property
    def on_hover(self) -> OptionalControlEventCallable:
        return self._get_event_handler("hover")

    @on_hover.setter
    def on_hover(self, handler: OptionalControlEventCallable):
        self._add_event_handler("hover", handler)
        self._set_attr("onHover", True if handler is not None else None)

    # on_focus
    @property
    def on_focus(self) -> OptionalControlEventCallable:
        return self._get_event_handler("focus")

    @on_focus.setter
    def on_focus(self, handler: OptionalControlEventCallable):
        self._add_event_handler("focus", handler)

    # on_blur
    @property
    def on_blur(self) -> OptionalControlEventCallable:
        return self._get_event_handler("blur")

    @on_blur.setter
    def on_blur(self, handler: OptionalControlEventCallable):
        self._add_event_handler("blur", handler)



import json
from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional

from flet.core.control import Control
from flet.core.control_event import ControlEvent
from flet.core.event_handler import EventHandler
from flet.core.ref import Ref
from flet.core.types import (
    ColorValue,
    DurationValue,
    OptionalControlEventCallable,
    OptionalEventCallable,
    OptionalNumber,
)
from flet.utils import deprecated


class GeolocatorPositionAccuracy(Enum):
    LOWEST = "lowest"
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    BEST = "best"
    BEST_FOR_NAVIGATION = "bestForNavigation"
    REDUCED = "reduced"


class GeolocatorPermissionStatus(Enum):
    DENIED = "denied"
    DENIED_FOREVER = "deniedForever"
    WHILE_IN_USE = "whileInUse"
    ALWAYS = "always"
    UNABLE_TO_DETERMINE = "unableToDetermine"


class GeolocatorActivityType(Enum):
    AUTOMOTIVE_NAVIGATION = "automotiveNavigation"
    FITNESS = "fitness"
    OTHER_NAVIGATION = "otherNavigation"
    AIRBORNE = "airborne"
    OTHER = "other"


@dataclass
class GeolocatorPosition:
    latitude: OptionalNumber = None
    longitude: OptionalNumber = None
    speed: OptionalNumber = None
    altitude: OptionalNumber = None
    timestamp: OptionalNumber = None
    accuracy: OptionalNumber = None
    altitude_accuracy: OptionalNumber = None
    heading: OptionalNumber = None
    heading_accuracy: OptionalNumber = None
    speed_accuracy: OptionalNumber = None
    floor: Optional[int] = None
    is_mocked: Optional[bool] = None


@dataclass
class GeolocatorSettings:
    accuracy: Optional[GeolocatorPositionAccuracy] = None
    distance_filter: Optional[int] = None
    time_limit: Optional[DurationValue] = None


@dataclass
class GeolocatorWebSettings(GeolocatorSettings):
    maximum_age: Optional[DurationValue] = None


@dataclass
class GeolocatorAppleSettings(GeolocatorSettings):
    activity_type: Optional[GeolocatorActivityType] = None
    pause_location_updates_automatically: Optional[bool] = False
    show_background_location_indicator: Optional[bool] = False
    allow_background_location_updates: Optional[bool] = True


@dataclass
class GeolocatorAndroidSettings(GeolocatorSettings):
    force_location_manager: Optional[bool] = False
    interval_duration: Optional[DurationValue] = False
    foreground_notification_text: Optional[str] = None
    foreground_notification_title: Optional[str] = None
    foreground_notification_channel_name: Optional[str] = "Background Location"
    foreground_notification_enable_wake_lock: Optional[bool] = False
    foreground_notification_enable_wifi_lock: Optional[bool] = False
    foreground_notification_set_ongoing: Optional[bool] = False
    foreground_notification_color: Optional[ColorValue] = None


class GeolocatorPositionChangeEvent(ControlEvent):
    def __init__(self, e: ControlEvent):
        super().__init__(e.target, e.name, e.data, e.control, e.page)
        d = json.loads(e.data)
        self.latitude: float = d.get("lat")
        self.longitude: float = d.get("long")


@deprecated(
    reason="Geolocator control has been moved to a separate Python package: https://pypi.org/project/flet-geolocator. "
    + "Read more about this change in Flet blog: https://flet.dev/blog/flet-v-0-26-release-announcement",
    version="0.26.0",
    delete_version="0.29.0",
)
class Geolocator(Control):
    """
    A control that allows you to fetch GPS data from your device.
    This control is non-visual and should be added to `page.overlay` list.

    -----

    Online docs: https://flet.dev/docs/controls/geolocator
    """

    def __init__(
        self,
        # Control
        #
        ref: Optional[Ref] = None,
        data: Any = None,
        location_settings: Optional[GeolocatorSettings] = None,
        on_position_change: OptionalEventCallable[GeolocatorPositionChangeEvent] = None,
        on_error: OptionalControlEventCallable = None,
    ):
        Control.__init__(
            self,
            ref=ref,
            data=data,
        )
        self.__on_position_change = EventHandler(
            lambda e: GeolocatorPositionChangeEvent(e)
        )
        self._add_event_handler(
            "positionChange", self.__on_position_change.get_handler()
        )
        self.on_position_change = on_position_change
        self.on_error = on_error
        self.location_settings = location_settings

    def _get_control_name(self):
        return "geolocator"

    def before_update(self):
        self._set_attr_json("locationSettings", self.location_settings)

    def get_current_position(
        self,
        accuracy: Optional[
            GeolocatorPositionAccuracy
        ] = GeolocatorPositionAccuracy.BEST,
        location_settings: Optional[GeolocatorSettings] = None,
        wait_timeout: Optional[float] = 25,
    ) -> GeolocatorPosition:
        ls = (
            location_settings
            or self.location_settings
            or GeolocatorSettings(accuracy=accuracy)
        )
        output = self.invoke_method(
            "get_current_position",
            {"location_settings": self._convert_attr_json(ls)},
            wait_for_result=True,
            wait_timeout=wait_timeout,
        )
        return (
            GeolocatorPosition(**json.loads(output))
            if output is not None
            else GeolocatorPosition()
        )

    async def get_current_position_async(
        self,
        accuracy: Optional[
            GeolocatorPositionAccuracy
        ] = GeolocatorPositionAccuracy.BEST,
        location_settings: Optional[GeolocatorSettings] = None,
        wait_timeout: Optional[float] = 25,
    ) -> GeolocatorPosition:
        ls = (
            location_settings
            or self.location_settings
            or GeolocatorSettings(accuracy=accuracy)
        )
        output = await self.invoke_method_async(
            "get_current_position",
            {"location_settings": self._convert_attr_json(ls)},
            wait_for_result=True,
            wait_timeout=wait_timeout,
        )
        return (
            GeolocatorPosition(**json.loads(output))
            if output is not None
            else GeolocatorPosition()
        )

    def get_last_known_position(
        self,
        wait_timeout: Optional[float] = 25,
    ) -> GeolocatorPosition:
        assert not self.page.web, "get_last_known_position is not supported on web"
        output = self.invoke_method(
            "get_last_known_position",
            wait_for_result=True,
            wait_timeout=wait_timeout,
        )
        return (
            GeolocatorPosition(**json.loads(output))
            if output is not None
            else GeolocatorPosition()
        )

    async def get_last_known_position_async(
        self,
        wait_timeout: Optional[float] = 25,
    ) -> GeolocatorPosition:
        assert not self.page.web, "get_last_known_position is not supported on web"
        output = await self.invoke_method_async(
            "get_last_known_position",
            wait_for_result=True,
            wait_timeout=wait_timeout,
        )
        return (
            GeolocatorPosition(**json.loads(output))
            if output is not None
            else GeolocatorPosition()
        )

    def get_permission_status(
        self, wait_timeout: Optional[float] = 25
    ) -> GeolocatorPermissionStatus:
        p = self.invoke_method(
            "get_permission_status",
            wait_for_result=True,
            wait_timeout=wait_timeout,
        )
        return GeolocatorPermissionStatus(p)

    async def get_permission_status_async(
        self, wait_timeout: Optional[float] = 25
    ) -> GeolocatorPermissionStatus:
        p = await self.invoke_method_async(
            "get_permission_status",
            wait_for_result=True,
            wait_timeout=wait_timeout,
        )
        return GeolocatorPermissionStatus(p)

    def request_permission(
        self, wait_timeout: Optional[float] = 25
    ) -> GeolocatorPermissionStatus:
        p = self.invoke_method(
            "request_permission",
            wait_for_result=True,
            wait_timeout=wait_timeout,
        )
        return GeolocatorPermissionStatus(p)

    async def request_permission_async(
        self, wait_timeout: Optional[float] = 25
    ) -> GeolocatorPermissionStatus:
        p = await self.invoke_method_async(
            "request_permission",
            wait_for_result=True,
            wait_timeout=wait_timeout,
        )
        return GeolocatorPermissionStatus(p)

    def is_location_service_enabled(self, wait_timeout: Optional[float] = 10) -> bool:
        enabled = self.invoke_method(
            "is_location_service_enabled",
            wait_for_result=True,
            wait_timeout=wait_timeout,
        )
        return enabled == "true"

    async def is_location_service_enabled_async(
        self, wait_timeout: Optional[float] = 10
    ) -> bool:
        enabled = await self.invoke_method_async(
            "is_location_service_enabled",
            wait_for_result=True,
            wait_timeout=wait_timeout,
        )
        return enabled == "true"

    def open_app_settings(self, wait_timeout: Optional[float] = 10) -> bool:
        assert not self.page.web, "open_app_settings is not supported on web"
        opened = self.invoke_method(
            "open_app_settings",
            wait_for_result=True,
            wait_timeout=wait_timeout,
        )
        return opened == "true"

    async def open_app_settings_async(self, wait_timeout: Optional[float] = 10) -> bool:
        assert not self.page.web, "open_app_settings is not supported on web"
        opened = await self.invoke_method_async(
            "open_app_settings",
            wait_for_result=True,
            wait_timeout=wait_timeout,
        )
        return opened == "true"

    def open_location_settings(self, wait_timeout: Optional[float] = 10) -> bool:
        assert not self.page.web, "open_location_settings is not supported on web"
        opened = self.invoke_method(
            "open_location_settings",
            wait_for_result=True,
            wait_timeout=wait_timeout,
        )
        return opened == "true"

    async def open_location_settings_async(
        self, wait_timeout: Optional[float] = 10
    ) -> bool:
        assert not self.page.web, "open_location_settings is not supported on web"
        opened = await self.invoke_method_async(
            "open_location_settings",
            wait_for_result=True,
            wait_timeout=wait_timeout,
        )
        return opened == "true"

    @property
    def on_position_change(
        self,
    ) -> OptionalEventCallable[GeolocatorPositionChangeEvent]:
        return self.__on_position_change.handler

    @on_position_change.setter
    def on_position_change(
        self, handler: OptionalEventCallable[GeolocatorPositionChangeEvent]
    ):
        self.__on_position_change.handler = handler
        self._set_attr("onPositionChange", True if handler is not None else None)

    @property
    def on_error(self) -> OptionalControlEventCallable:
        return self._get_attr("error")

    @on_error.setter
    def on_error(self, handler: OptionalControlEventCallable):
        self._add_event_handler("error", handler)



from typing import Any, List, Optional

from flet.core.adaptive_control import AdaptiveControl
from flet.core.control import Control
from flet.core.ref import Ref


class CupertinoContextMenu(AdaptiveControl):
    """
    A full-screen modal route that opens up when the content is long-pressed.

    -----

    Online docs: https://flet.dev/docs/controls/cupertinocontextmenu
    """

    def __init__(
        self,
        content: Control,
        actions: List[Control],
        enable_haptic_feedback: Optional[bool] = None,
        #
        # Control
        #
        ref: Optional[Ref] = None,
        disabled: Optional[bool] = None,
        visible: Optional[bool] = None,
        data: Any = None,
        adaptive: Optional[bool] = False,
    ):
        Control.__init__(
            self,
            ref=ref,
            disabled=disabled,
            visible=visible,
            data=data,
        )

        AdaptiveControl.__init__(self, adaptive=adaptive)

        self.enable_haptic_feedback = enable_haptic_feedback
        self.content = content
        self.actions = actions

    def _get_control_name(self):
        return "cupertinocontextmenu"

    def _get_children(self):
        self.__content._set_attr_internal("n", "content")
        for action in self.__actions:
            action._set_attr_internal("n", "action")
        return [self.__content] + self.__actions

    def before_update(self):
        super().before_update()
        assert (
            len(self.__actions) > 0
        ), "actions must be provided and at least one must be visible"

    # enable_haptic_feedback
    @property
    def enable_haptic_feedback(self) -> bool:
        return self._get_attr("enableHapticFeedback", data_type="bool", def_value=False)

    @enable_haptic_feedback.setter
    def enable_haptic_feedback(self, value: Optional[bool]):
        self._set_attr("enableHapticFeedback", value)

    # content
    @property
    def content(self) -> Control:
        return self.__content

    @content.setter
    def content(self, value: Control):
        self.__content = value

    # actions
    @property
    def actions(self) -> List[Control]:
        return self.__actions

    @actions.setter
    def actions(self, value: List[Control]):
        self.__actions = value



from dataclasses import dataclass
from enum import Enum
from typing import Optional, Union

from flet.core.types import DurationValue
from flet.utils import deprecated

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class AnimationCurve(Enum):
    BOUNCE_IN = "bounceIn"
    BOUNCE_IN_OUT = "bounceInOut"
    BOUNCE_OUT = "bounceOut"
    DECELERATE = "decelerate"
    EASE = "ease"
    EASE_IN = "easeIn"
    EASE_IN_BACK = "easeInBack"
    EASE_IN_CIRC = "easeInCirc"
    EASE_IN_CUBIC = "easeInCubic"
    EASE_IN_EXPO = "easeInExpo"
    EASE_IN_OUT = "easeInOut"
    EASE_IN_OUT_BACK = "easeInOutBack"
    EASE_IN_OUT_CIRC = "easeInOutCirc"
    EASE_IN_OUT_CUBIC = "easeInOutCubic"
    EASE_IN_OUT_CUBIC_EMPHASIZED = "easeInOutCubicEmphasized"
    EASE_IN_OUT_EXPO = "easeInOutExpo"
    EASE_IN_OUT_QUAD = "easeInOutQuad"
    EASE_IN_OUT_QUART = "easeInOutQuart"
    EASE_IN_OUT_QUINT = "easeInOutQuint"
    EASE_IN_OUT_SINE = "easeInOutSine"
    EASE_IN_QUAD = "easeInQuad"
    EASE_IN_QUART = "easeInQuart"
    EASE_IN_QUINT = "easeInQuint"
    EASE_IN_SINE = "easeInSine"
    EASE_IN_TO_LINEAR = "easeInToLinear"
    EASE_OUT = "easeOut"
    EASE_OUT_BACK = "easeOutBack"
    EASE_OUT_CIRC = "easeOutCirc"
    EASE_OUT_CUBIC = "easeOutCubic"
    EASE_OUT_EXPO = "easeOutExpo"
    EASE_OUT_QUAD = "easeOutQuad"
    EASE_OUT_QUART = "easeOutQuart"
    EASE_OUT_QUINT = "easeOutQuint"
    EASE_OUT_SINE = "easeOutSine"
    ELASTIC_IN = "elasticIn"
    ELASTIC_IN_OUT = "elasticInOut"
    ELASTIC_OUT = "elasticOut"
    FAST_LINEAR_TO_SLOW_EASE_IN = "fastLinearToSlowEaseIn"
    FAST_OUT_SLOWIN = "fastOutSlowIn"
    LINEAR = "linear"
    LINEAR_TO_EASE_OUT = "linearToEaseOut"
    SLOW_MIDDLE = "slowMiddle"


@dataclass
class Animation:
    duration: DurationValue = None
    curve: Optional[AnimationCurve] = None


@dataclass
class AnimationStyle:
    duration: Optional[DurationValue] = None
    reverse_duration: Optional[DurationValue] = None
    curve: Optional[AnimationCurve] = None
    reverse_curve: Optional[AnimationCurve] = None


@deprecated(
    reason="Use Animation class instead.", version="0.26.0", delete_version="0.29.0"
)
def implicit(duration: DurationValue, curve: Optional[AnimationCurve] = None):
    return Animation(duration=duration, curve=curve)


AnimationValue = Union[bool, int, Animation]



import json
from dataclasses import dataclass, field
from typing import Any, List, Optional

from flet.core.control import Control, OptionalNumber
from flet.core.control_event import ControlEvent
from flet.core.event_handler import EventHandler
from flet.core.ref import Ref
from flet.core.types import OptionalEventCallable


@dataclass
class AutoCompleteSuggestion:
    key: str = field(default=None)
    value: str = field(default=None)


class AutoComplete(Control):
    """
    Helps the user make a selection by entering some text and choosing from among a list of displayed options.

    -----

    Online docs: https://flet.dev/docs/controls/autocomplete
    """

    def __init__(
        self,
        suggestions: Optional[List[AutoCompleteSuggestion]] = None,
        suggestions_max_height: OptionalNumber = None,
        on_select: OptionalEventCallable["AutoCompleteSelectEvent"] = None,
        #
        # Control
        #
        ref: Optional[Ref] = None,
        opacity: OptionalNumber = None,
        visible: Optional[bool] = None,
        data: Any = None,
    ):

        Control.__init__(
            self,
            ref=ref,
            opacity=opacity,
            visible=visible,
            data=data,
        )

        self.__on_select = EventHandler(lambda e: AutoCompleteSelectEvent(e))
        self._add_event_handler("select", self.__on_select.get_handler())

        self.suggestions = suggestions
        self.suggestions_max_height = suggestions_max_height
        self.on_select = on_select

    def _get_control_name(self):
        return "autocomplete"

    def before_update(self):
        self._set_attr_json("suggestions", self.__suggestions)

    # selected_index
    @property
    def selected_index(self) -> Optional[int]:
        return self._get_attr("selectedIndex", data_type="int")

    # suggestions_max_height
    @property
    def suggestions_max_height(self) -> float:
        return self._get_attr(
            "suggestionsMaxHeight", data_type="float", def_value=200.0
        )

    @suggestions_max_height.setter
    def suggestions_max_height(self, value: OptionalNumber):
        assert value is None or value >= 0, "suggestions_max_height cannot be negative"
        self._set_attr("suggestionsMaxHeight", value)

    # suggestions
    @property
    def suggestions(self) -> Optional[List[AutoCompleteSuggestion]]:
        return self.__suggestions

    @suggestions.setter
    def suggestions(self, value: Optional[List[str]]):
        self.__suggestions = value or []

    # on_select
    @property
    def on_select(self) -> OptionalEventCallable["AutoCompleteSelectEvent"]:
        return self.__on_select.handler

    @on_select.setter
    def on_select(self, handler: OptionalEventCallable["AutoCompleteSelectEvent"]):
        self.__on_select.handler = handler


class AutoCompleteSelectEvent(ControlEvent):
    def __init__(self, e: ControlEvent):
        super().__init__(e.target, e.name, e.data, e.control, e.page)
        d = json.loads(e.data)
        self.selection: AutoCompleteSuggestion = AutoCompleteSuggestion(
            key=d.get("key"), value=d.get("value")
        )



from enum import Enum
from typing import Any, Optional, Union

from flet.core.alignment import Alignment
from flet.core.animation import AnimationValue
from flet.core.badge import BadgeValue
from flet.core.constrained_control import ConstrainedControl
from flet.core.control import OptionalNumber
from flet.core.ref import Ref
from flet.core.tooltip import TooltipValue
from flet.core.types import (
    ColorEnums,
    ColorValue,
    OffsetValue,
    OptionalControlEventCallable,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
)


class CupertinoTimerPickerMode(Enum):
    HOUR_MINUTE = "hm"
    HOUR_MINUTE_SECONDS = "hms"
    MINUTE_SECONDS = "ms"


class CupertinoTimerPicker(ConstrainedControl):
    """
    A countdown timer picker in iOS style.

    It can show a countdown duration with hour, minute and second spinners. The duration is bound between 0 and 23 hours 59 minutes 59 seconds.

    -----

    Online docs: https://flet.dev/docs/controls/cupertinotimerpicker
    """

    def __init__(
        self,
        value: Optional[int] = None,
        alignment: Optional[Alignment] = None,
        second_interval: OptionalNumber = None,
        minute_interval: OptionalNumber = None,
        mode: Optional[CupertinoTimerPickerMode] = None,
        bgcolor: Optional[ColorValue] = None,
        item_extent: OptionalNumber = None,
        on_change: OptionalControlEventCallable = None,
        #
        # ConstrainedControl
        #
        ref: Optional[Ref] = None,
        key: Optional[str] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        left: OptionalNumber = None,
        top: OptionalNumber = None,
        right: OptionalNumber = None,
        bottom: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: OptionalControlEventCallable = None,
        tooltip: Optional[TooltipValue] = None,
        badge: Optional[BadgeValue] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        ConstrainedControl.__init__(
            self,
            ref=ref,
            key=key,
            width=width,
            height=height,
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            tooltip=tooltip,
            badge=badge,
            visible=visible,
            disabled=disabled,
            data=data,
        )

        self.value = value
        self.alignment = alignment
        self.mode = mode
        self.bgcolor = bgcolor
        self.on_change = on_change
        self.second_interval = second_interval
        self.minute_interval = minute_interval
        self.item_extent = item_extent

    def _get_control_name(self):
        return "cupertinotimerpicker"

    def before_update(self):
        super().before_update()
        self._set_attr_json("alignment", self.__alignment)

    # value
    @property
    def value(self) -> int:
        return self._get_attr("value", data_type="int", def_value=0)

    @value.setter
    def value(self, value: Optional[int]):
        self._set_attr("value", value)

    # alignment
    @property
    def alignment(self) -> Optional[Alignment]:
        return self.__alignment

    @alignment.setter
    def alignment(self, value: Optional[Alignment]):
        self.__alignment = value

    # bgcolor
    @property
    def bgcolor(self) -> Optional[ColorValue]:
        return self.__bgcolor

    @bgcolor.setter
    def bgcolor(self, value: Optional[ColorValue]):
        self.__bgcolor = value
        self._set_enum_attr("bgcolor", value, ColorEnums)

    # second_interval
    @property
    def second_interval(self) -> int:
        return self._get_attr("secondInterval", data_type="int", def_value=1)

    @second_interval.setter
    def second_interval(self, value: OptionalNumber):
        self._set_attr("secondInterval", value)

    # item_extent
    @property
    def item_extent(self) -> float:
        return self._get_attr("itemExtent", data_type="float", def_value=32.0)

    @item_extent.setter
    def item_extent(self, value: OptionalNumber):
        self._set_attr("itemExtent", value)

    # minute_interval
    @property
    def minute_interval(self) -> int:
        return self._get_attr("minuteInterval", data_type="int", def_value=1)

    @minute_interval.setter
    def minute_interval(self, value: OptionalNumber):
        self._set_attr("minuteInterval", value)

    # mode
    @property
    def mode(self) -> Optional[CupertinoTimerPickerMode]:
        return self.__mode

    @mode.setter
    def mode(self, value: Optional[CupertinoTimerPickerMode]):
        self.__mode = value
        self._set_enum_attr("mode", value, CupertinoTimerPickerMode)

    # on_change
    @property
    def on_change(self) -> OptionalControlEventCallable:
        return self._get_event_handler("change")

    @on_change.setter
    def on_change(self, handler: OptionalControlEventCallable):
        self._add_event_handler("change", handler)



import json
from typing import Any, List, Optional, Tuple, Union

from flet.core.adaptive_control import AdaptiveControl
from flet.core.alignment import Alignment
from flet.core.animation import AnimationValue
from flet.core.badge import BadgeValue
from flet.core.blur import Blur
from flet.core.border import Border
from flet.core.box import (
    BoxDecoration,
    BoxShadow,
    BoxShape,
    ColorFilter,
    DecorationImage,
)
from flet.core.constrained_control import ConstrainedControl
from flet.core.control import Control, OptionalNumber
from flet.core.control_event import ControlEvent
from flet.core.event_handler import EventHandler
from flet.core.gradients import Gradient
from flet.core.ref import Ref
from flet.core.theme import Theme
from flet.core.tooltip import TooltipValue
from flet.core.types import (
    BlendMode,
    BorderRadiusValue,
    ClipBehavior,
    ColorEnums,
    ColorValue,
    ImageFit,
    ImageRepeat,
    MarginValue,
    OffsetValue,
    OptionalControlEventCallable,
    OptionalEventCallable,
    PaddingValue,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
    ThemeMode,
    UrlTarget,
)
from flet.utils.deprecated import deprecated_property


class Container(ConstrainedControl, AdaptiveControl):
    """
    Container allows to decorate a control with background color and border and position it with padding, margin and alignment.

    Example:

    ```
    import flet as ft

    def main(page: ft.Page):
        page.title = "Container"

        c1 = ft.Container(
            content=ft.Text("Container with background"),
            bgcolor=ft.colors.AMBER_100,
            padding=5,
        )
        page.add(c1)

    ft.app(target=main)
    ```

    -----

    Online docs: https://flet.dev/docs/controls/container
    """

    def __init__(
        self,
        content: Optional[Control] = None,
        padding: Optional[PaddingValue] = None,
        margin: Optional[MarginValue] = None,
        alignment: Optional[Alignment] = None,
        bgcolor: Optional[ColorValue] = None,
        gradient: Optional[Gradient] = None,
        blend_mode: Optional[BlendMode] = None,
        border: Optional[Border] = None,
        border_radius: Optional[BorderRadiusValue] = None,
        shape: Optional[BoxShape] = None,
        clip_behavior: Optional[ClipBehavior] = None,
        ink: Optional[bool] = None,
        image: Optional[DecorationImage] = None,
        ink_color: Optional[ColorValue] = None,
        animate: Optional[AnimationValue] = None,
        blur: Union[
            None, float, int, Tuple[Union[float, int], Union[float, int]], Blur
        ] = None,
        shadow: Union[None, BoxShadow, List[BoxShadow]] = None,
        url: Optional[str] = None,
        url_target: Optional[UrlTarget] = None,
        theme: Optional[Theme] = None,
        dark_theme: Optional[Theme] = None,
        theme_mode: Optional[ThemeMode] = None,
        color_filter: Optional[ColorFilter] = None,
        ignore_interactions: Optional[bool] = None,
        foreground_decoration: Optional[BoxDecoration] = None,
        on_click: OptionalControlEventCallable = None,
        on_tap_down: OptionalEventCallable["ContainerTapEvent"] = None,
        on_long_press: OptionalControlEventCallable = None,
        on_hover: OptionalControlEventCallable = None,
        #
        # ConstrainedControl and AdaptiveControl
        #
        ref: Optional[Ref] = None,
        key: Optional[str] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        left: OptionalNumber = None,
        top: OptionalNumber = None,
        right: OptionalNumber = None,
        bottom: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: OptionalControlEventCallable = None,
        tooltip: Optional[TooltipValue] = None,
        badge: Optional[BadgeValue] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
        rtl: Optional[bool] = None,
        adaptive: Optional[bool] = None,
    ):
        ConstrainedControl.__init__(
            self,
            ref=ref,
            key=key,
            width=width,
            height=height,
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            tooltip=tooltip,
            badge=badge,
            visible=visible,
            disabled=disabled,
            data=data,
            rtl=rtl,
        )

        AdaptiveControl.__init__(self, adaptive=adaptive)

        self.__on_tap_down = EventHandler(lambda e: ContainerTapEvent(e))
        self._add_event_handler("tap_down", self.__on_tap_down.get_handler())

        self.content = content
        self.padding = padding
        self.margin = margin
        self.alignment = alignment
        self.bgcolor = bgcolor
        self.gradient = gradient
        self.blend_mode = blend_mode
        self.border = border
        self.border_radius = border_radius
        self.shape = shape
        self.clip_behavior = clip_behavior
        self.ink = ink
        self.ink_color = ink_color
        self.animate = animate
        self.blur = blur
        self.shadow = shadow
        self.url = url
        self.url_target = url_target
        self.theme = theme
        self.dark_theme = dark_theme
        self.theme_mode = theme_mode
        self.color_filter = color_filter
        self.ignore_interactions = ignore_interactions
        self.on_click = on_click
        self.on_tap_down = on_tap_down
        self.on_long_press = on_long_press
        self.on_hover = on_hover
        self.image = image
        self.foreground_decoration = foreground_decoration

    def _get_control_name(self):
        return "container"

    def before_update(self):
        super().before_update()
        assert (
            self.__blend_mode is None
            or self.__gradient is not None
            or self.bgcolor is not None
        ), "blend_mode applies to bgcolor or gradient, but no bgcolor or gradient was provided"
        assert (
            self.__shape != BoxShape.CIRCLE or self.__border_radius is None
        ), "border_radius is not supported with shape=BoxShape.CIRCLE"
        self._set_attr_json("borderRadius", self.__border_radius)
        self._set_attr_json("border", self.__border)
        self._set_attr_json("margin", self.__margin)
        self._set_attr_json("padding", self.__padding)
        self._set_attr_json("alignment", self.__alignment)
        self._set_attr_json("gradient", self.__gradient)
        self._set_attr_json("animate", self.__animate)
        self._set_attr_json("blur", self.__blur)
        self._set_attr_json("shadow", self.__shadow if self.__shadow else None)
        self._set_attr_json("theme", self.__theme)
        self._set_attr_json("darkTheme", self.__dark_theme)
        self._set_attr_json("colorFilter", self.__color_filter)
        self._set_attr_json("image", self.__image)
        self._set_attr_json("foregroundDecoration", self.__foreground_decoration)

    def _get_children(self):
        children = []
        if self.__content is not None:
            self.__content._set_attr_internal("n", "content")
            children.append(self.__content)
        return children

    # alignment
    @property
    def alignment(self) -> Optional[Alignment]:
        """:obj:`Alignment`, optional: Align the child control within the container.

        Alignment is an instance of `alignment.Alignment` class object with `x` and `y` properties
        representing the distance from the center of a rectangle.
        """
        return self.__alignment

    @alignment.setter
    def alignment(self, value: Optional[Alignment]):
        self.__alignment = value

    # padding
    @property
    def padding(self) -> Optional[PaddingValue]:
        return self.__padding

    @padding.setter
    def padding(self, value: Optional[PaddingValue]):
        self.__padding = value

    # image
    @property
    def image(self) -> Optional[DecorationImage]:
        return self.__image

    @image.setter
    def image(self, value: Optional[DecorationImage]):
        self.__image = value

    # foreground_decoration
    @property
    def foreground_decoration(self) -> Optional[BoxDecoration]:
        return self.__foreground_decoration

    @foreground_decoration.setter
    def foreground_decoration(self, value: Optional[BoxDecoration]):
        self.__foreground_decoration = value

    # margin
    @property
    def margin(self) -> Optional[MarginValue]:
        return self.__margin

    @margin.setter
    def margin(self, value: Optional[MarginValue]):
        self.__margin = value

    # bgcolor
    @property
    def bgcolor(self):
        return self.__bgcolor

    @bgcolor.setter
    def bgcolor(self, value: Optional[ColorValue]):
        self.__bgcolor = value
        self._set_enum_attr("bgColor", value, ColorEnums)

    # gradient
    @property
    def gradient(self) -> Optional[Gradient]:
        return self.__gradient

    @gradient.setter
    def gradient(self, value: Optional[Gradient]):
        self.__gradient = value

    # blend_mode
    @property
    def blend_mode(self) -> Optional[BlendMode]:
        return self.__blend_mode

    @blend_mode.setter
    def blend_mode(self, value: Optional[BlendMode]):
        self.__blend_mode = value
        self._set_enum_attr("blendMode", value, BlendMode)

    # blur
    @property
    def blur(
        self,
    ) -> Union[None, float, int, Tuple[Union[float, int], Union[float, int]], Blur]:
        return self.__blur

    @blur.setter
    def blur(
        self,
        value: Union[
            None, float, int, Tuple[Union[float, int], Union[float, int]], Blur
        ],
    ):
        self.__blur = value

    # shadow
    @property
    def shadow(self) -> Union[None, BoxShadow, List[BoxShadow]]:
        return self.__shadow

    @shadow.setter
    def shadow(self, value: Union[None, BoxShadow, List[BoxShadow]]):
        self.__shadow = value if value is not None else []

    # color_filter
    @property
    def color_filter(self) -> Optional[ColorFilter]:
        return self.__color_filter

    @color_filter.setter
    def color_filter(self, value: Optional[ColorFilter]):
        self.__color_filter = value

    # border
    @property
    def border(self) -> Optional[Border]:
        return self.__border

    @border.setter
    def border(self, value: Optional[Border]):
        self.__border = value

    # border_radius
    @property
    def border_radius(self) -> Optional[BorderRadiusValue]:
        return self.__border_radius

    @border_radius.setter
    def border_radius(self, value: Optional[BorderRadiusValue]):
        self.__border_radius = value

    # ignore_interactions
    @property
    def ignore_interactions(self) -> Optional[bool]:
        return self._get_attr("ignoreInteractions", data_type="bool", def_value=False)

    @ignore_interactions.setter
    def ignore_interactions(self, value: Optional[str]):
        self._set_attr("ignoreInteractions", value)

    # content
    @property
    def content(self) -> Optional[Control]:
        return self.__content

    @content.setter
    def content(self, value: Optional[Control]):
        self.__content = value

    # shape
    @property
    def shape(self) -> Optional[BoxShape]:
        return self.__shape

    @shape.setter
    def shape(self, value: Optional[BoxShape]):
        self.__shape = value
        self._set_enum_attr("shape", value, BoxShape)

    # clip_behavior
    @property
    def clip_behavior(self) -> Optional[ClipBehavior]:
        return self.__clip_behavior

    @clip_behavior.setter
    def clip_behavior(self, value: Optional[ClipBehavior]):
        self.__clip_behavior = value
        self._set_enum_attr("clipBehavior", value, ClipBehavior)

    # ink
    @property
    def ink(self) -> bool:
        return self._get_attr("ink", data_type="bool", def_value=False)

    @ink.setter
    def ink(self, value: Optional[bool]):
        self._set_attr("ink", value)

    # ink color
    @property
    def ink_color(self) -> Optional[ColorValue]:
        return self.__ink_color

    @ink_color.setter
    def ink_color(self, value: Optional[ColorValue]):
        self.__ink_color = value
        self._set_enum_attr("inkColor", value, ColorEnums)

    # animate
    @property
    def animate(self) -> Optional[AnimationValue]:
        return self.__animate

    @animate.setter
    def animate(self, value: Optional[AnimationValue]):
        self.__animate = value

    # url
    @property
    def url(self) -> Optional[str]:
        return self._get_attr("url")

    @url.setter
    def url(self, value: Optional[str]):
        self._set_attr("url", value)

    # url_target
    @property
    def url_target(self) -> Optional[UrlTarget]:
        return self.__url_target

    @url_target.setter
    def url_target(self, value: Optional[UrlTarget]):
        self.__url_target = value
        self._set_enum_attr("urlTarget", value, UrlTarget)

    # theme
    @property
    def theme(self) -> Optional[Theme]:
        return self.__theme

    @theme.setter
    def theme(self, value: Optional[Theme]):
        self.__theme = value

    # dark_theme
    @property
    def dark_theme(self) -> Optional[Theme]:
        return self.__dark_theme

    @dark_theme.setter
    def dark_theme(self, value: Optional[Theme]):
        self.__dark_theme = value

    # theme_mode
    @property
    def theme_mode(self) -> Optional[ThemeMode]:
        return self.__theme_mode

    @theme_mode.setter
    def theme_mode(self, value: Optional[ThemeMode]):
        self.__theme_mode = value
        self._set_enum_attr("themeMode", value, ThemeMode)

    # on_click
    @property
    def on_click(self):
        return self._get_event_handler("click")

    @on_click.setter
    def on_click(self, handler: OptionalControlEventCallable):
        self._add_event_handler("click", handler)
        self._set_attr("onClick", True if handler is not None else None)

    # on_tap_down
    @property
    def on_tap_down(self) -> OptionalEventCallable["ContainerTapEvent"]:
        return self.__on_tap_down.handler

    @on_tap_down.setter
    def on_tap_down(self, handler: OptionalEventCallable["ContainerTapEvent"]):
        self.__on_tap_down.handler = handler
        self._set_attr("onTapDown", True if handler is not None else None)

    # on_long_press
    @property
    def on_long_press(self):
        return self._get_event_handler("long_press")

    @on_long_press.setter
    def on_long_press(self, handler: OptionalControlEventCallable):
        self._add_event_handler("long_press", handler)
        self._set_attr("onLongPress", True if handler is not None else None)

    # on_hover
    @property
    def on_hover(self):
        return self._get_event_handler("hover")

    @on_hover.setter
    def on_hover(self, handler: OptionalControlEventCallable):
        self._add_event_handler("hover", handler)
        self._set_attr("onHover", True if handler is not None else None)


class ContainerTapEvent(ControlEvent):
    def __init__(self, e: ControlEvent):
        super().__init__(e.target, e.name, e.data, e.control, e.page)
        d = json.loads(e.data)
        self.local_x: float = d.get("lx")
        self.local_y: float = d.get("ly")
        self.global_x: float = d.get("gx")
        self.global_y: float = d.get("gy")



import dataclasses
import math
from dataclasses import field
from enum import Enum
from typing import List, Optional, Union

from flet.core import alignment
from flet.core.alignment import Alignment


class GradientTileMode(Enum):
    CLAMP = "clamp"
    DECAL = "decal"
    MIRROR = "mirror"
    REPEATED = "repeated"


@dataclasses.dataclass
class Gradient:
    colors: List[str]
    tile_mode: GradientTileMode = field(default=GradientTileMode.CLAMP)
    rotation: Union[None, float, int] = field(default=None)
    stops: Optional[List[float]] = field(default=None)


@dataclasses.dataclass
class LinearGradient(Gradient):
    begin: Alignment = field(default_factory=lambda: alignment.center_left)
    end: Alignment = field(default_factory=lambda: alignment.center_right)
    type: str = field(default="linear")


@dataclasses.dataclass
class RadialGradient(Gradient):
    center: Alignment = field(default_factory=lambda: alignment.center)
    radius: Union[float, int] = field(default=0.5)
    focal: Optional[Alignment] = field(default=None)
    focal_radius: Union[float, int] = field(default=0.0)
    type: str = field(default="radial")


@dataclasses.dataclass
class SweepGradient(Gradient):
    center: Alignment = field(default_factory=lambda: alignment.center)
    start_angle: Union[float, int] = field(default=0.0)
    end_angle: Union[float, int] = field(default=math.pi * 2)
    type: str = field(default="sweep")



import asyncio
import json
import secrets
import threading
import time
from typing import List, Optional, Tuple

import httpx
from oauthlib.oauth2 import WebApplicationClient
from oauthlib.oauth2.rfc6749.tokens import OAuth2Token

from flet.auth.oauth_provider import OAuthProvider
from flet.auth.oauth_token import OAuthToken
from flet.auth.user import User
from flet.core.locks import AsyncNopeLock, NopeLock
from flet.utils import is_asyncio
from flet.version import version


class Authorization:
    def __init__(
        self,
        provider: OAuthProvider,
        fetch_user: bool,
        fetch_groups: bool,
        scope: Optional[List[str]] = None,
    ) -> None:
        self.fetch_user = fetch_user
        self.fetch_groups = fetch_groups
        self.scope = scope if scope is not None else []
        self.provider = provider
        self.__token: Optional[OAuthToken] = None
        self.user: Optional[User] = None
        self.__lock = threading.Lock() if not is_asyncio() else NopeLock()
        self.__async_lock = asyncio.Lock() if is_asyncio() else AsyncNopeLock()

        # fix scopes
        self.scope.extend(self.provider.scopes)
        if self.fetch_user:
            for s in self.provider.user_scopes:
                if s not in self.scope:
                    self.scope.append(s)
        if self.fetch_groups:
            for s in self.provider.group_scopes:
                if s not in self.scope:
                    self.scope.append(s)

    def dehydrate_token(self, saved_token: str):
        self.__token = OAuthToken.from_json(saved_token)
        self.__refresh_token()
        self.__fetch_user_and_groups()

    async def dehydrate_token_async(self, saved_token: str):
        self.__token = OAuthToken.from_json(saved_token)
        await self.__refresh_token_async()
        await self.__fetch_user_and_groups_async()

    # token
    @property
    def token(self) -> Optional[OAuthToken]:
        with self.__lock:
            self.__refresh_token()
            return self.__token

    # token_async
    @property
    async def token_async(self) -> Optional[OAuthToken]:
        async with self.__async_lock:
            await self.__refresh_token_async()
            return self.__token

    def get_authorization_data(self) -> Tuple[str, str]:
        self.state = secrets.token_urlsafe(16)
        client = WebApplicationClient(self.provider.client_id)
        authorization_url = client.prepare_request_uri(
            self.provider.authorization_endpoint,
            self.provider.redirect_url,
            scope=self.scope,
            state=self.state,
            code_challenge=self.provider.code_challenge,
            code_challenge_method=self.provider.code_challenge_method,
        )
        return authorization_url, self.state

    def request_token(self, code: str):
        req = self.__get_request_token_request(code)
        with httpx.Client(follow_redirects=True) as client:
            resp = client.send(req)
            resp.raise_for_status()
            client = WebApplicationClient(self.provider.client_id)
            t = client.parse_request_body_response(resp.text)
            self.__token = self.__convert_token(t)
            self.__fetch_user_and_groups()

    async def request_token_async(self, code: str):
        req = self.__get_request_token_request(code)
        async with httpx.AsyncClient(follow_redirects=True) as client:
            resp = await client.send(req)
            resp.raise_for_status()
            client = WebApplicationClient(self.provider.client_id)
            t = client.parse_request_body_response(resp.text)
            self.__token = self.__convert_token(t)
            await self.__fetch_user_and_groups_async()

    def __get_request_token_request(self, code: str):
        client = WebApplicationClient(self.provider.client_id)
        data = client.prepare_request_body(
            code=code,
            redirect_uri=self.provider.redirect_url,
            client_secret=self.provider.client_secret,
            include_client_id=True,
            code_verifier=self.provider.code_verifier,
        )
        headers = self.__get_default_headers()
        headers["content-type"] = "application/x-www-form-urlencoded"
        return httpx.Request(
            "POST", self.provider.token_endpoint, content=data, headers=headers
        )

    def __fetch_user_and_groups(self):
        assert self.__token is not None
        if self.fetch_user:
            self.user = self.provider._fetch_user(self.__token.access_token)
            if self.user is None and self.provider.user_endpoint is not None:
                if self.provider.user_id_fn is None:
                    raise Exception(
                        "user_id_fn must be specified too if user_endpoint is not None"
                    )
                self.user = self.__get_user()
            if self.fetch_groups and self.user is not None:
                self.user.groups = self.provider._fetch_groups(
                    self.__token.access_token
                )

    async def __fetch_user_and_groups_async(self):
        assert self.__token is not None
        if self.fetch_user:
            self.user = await self.provider._fetch_user_async(self.__token.access_token)
            if self.user is None and self.provider.user_endpoint is not None:
                if self.provider.user_id_fn is None:
                    raise Exception(
                        "user_id_fn must be specified too if user_endpoint is not None"
                    )
                self.user = await self.__get_user_async()
            if self.fetch_groups and self.user is not None:
                self.user.groups = await self.provider._fetch_groups_async(
                    self.__token.access_token
                )

    def __convert_token(self, t: OAuth2Token):
        return OAuthToken(
            access_token=t["access_token"],
            scope=t.get("scope"),
            token_type=t.get("token_type"),
            expires_in=t.get("expires_in"),
            expires_at=t.get("expires_at"),
            refresh_token=t.get("refresh_token"),
        )

    def __refresh_token(self):
        refresh_req = self.__get_refresh_token_request()
        if refresh_req:
            with httpx.Client(follow_redirects=True) as client:
                refresh_resp = client.send(refresh_req)
                self.__complete_refresh_token_request(refresh_resp)

    async def __refresh_token_async(self):
        refresh_req = self.__get_refresh_token_request()
        if refresh_req:
            async with httpx.AsyncClient(follow_redirects=True) as client:
                refresh_resp = await client.send(refresh_req)
                self.__complete_refresh_token_request(refresh_resp)

    def __get_refresh_token_request(self):
        if (
            self.__token is None
            or self.__token.expires_at is None
            or self.__token.refresh_token is None
            or time.time() < self.__token.expires_at
        ):
            return None

        assert self.__token is not None
        client = WebApplicationClient(self.provider.client_id)
        data = client.prepare_refresh_body(
            client_id=self.provider.client_id,
            client_secret=self.provider.client_secret,
            refresh_token=self.__token.refresh_token,
            redirect_uri=self.provider.redirect_url,
        )
        headers = self.__get_default_headers()
        headers["content-type"] = "application/x-www-form-urlencoded"
        return httpx.Request(
            "POST", url=self.provider.token_endpoint, content=data, headers=headers
        )

    def __complete_refresh_token_request(self, refresh_resp):
        refresh_resp.raise_for_status()
        assert self.__token is not None
        client = WebApplicationClient(self.provider.client_id)
        t = client.parse_request_body_response(refresh_resp.text)
        if t.get("refresh_token") is None:
            t["refresh_token"] = self.__token.refresh_token
        self.__token = self.__convert_token(t)

    def __get_user(self):
        user_req = self.__get_user_request()
        with httpx.Client() as client:
            user_resp = client.send(user_req)
            return self.__complete_user_request(user_resp)

    async def __get_user_async(self):
        user_req = self.__get_user_request()
        async with httpx.AsyncClient(follow_redirects=True) as client:
            user_resp = await client.send(user_req)
            return self.__complete_user_request(user_resp)

    def __get_user_request(self):
        assert self.token is not None
        assert self.provider.user_endpoint is not None
        headers = self.__get_default_headers()
        headers["Authorization"] = f"Bearer {self.token.access_token}"
        return httpx.Request("GET", self.provider.user_endpoint, headers=headers)

    def __complete_user_request(self, user_resp):
        user_resp.raise_for_status()
        assert self.provider.user_id_fn is not None
        uj = json.loads(user_resp.text)
        return User(uj, str(self.provider.user_id_fn(uj)))

    def __get_default_headers(self):
        return {
            "User-Agent": f"Flet/{version}",
        }



from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional, Union

from flet.core.alignment import Alignment
from flet.core.border import Border
from flet.core.gradients import Gradient
from flet.core.types import (
    BlendMode,
    BorderRadiusValue,
    ColorValue,
    ImageFit,
    ImageRepeat,
    Number,
    OffsetValue,
    OptionalNumber,
)


@dataclass
class ColorFilter:
    color: Optional[ColorValue] = None
    blend_mode: Optional[BlendMode] = None


class FilterQuality(Enum):
    NONE = "none"
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


class ShadowBlurStyle(Enum):
    NORMAL = "normal"
    SOLID = "solid"
    OUTER = "outer"
    INNER = "inner"


@dataclass
class BoxShadow:
    spread_radius: Optional[float] = None
    blur_radius: Optional[float] = None
    color: Optional[ColorValue] = None
    offset: Optional[OffsetValue] = None
    blur_style: ShadowBlurStyle = field(default=ShadowBlurStyle.NORMAL)


class BoxShape(Enum):
    RECTANGLE = "rectangle"
    CIRCLE = "circle"


@dataclass
class DecorationImage:
    src: Optional[str] = None
    src_base64: Optional[str] = None
    color_filter: Optional[ColorFilter] = None
    fit: Optional[ImageFit] = None
    alignment: Optional[Alignment] = None
    repeat: Optional[ImageRepeat] = None
    match_text_direction: Optional[bool] = None
    scale: OptionalNumber = None
    opacity: OptionalNumber = None
    filter_quality: Optional[FilterQuality] = None
    invert_colors: Optional[bool] = None
    anti_alias: Optional[bool] = None


@dataclass
class BoxDecoration:
    bgcolor: Optional[ColorValue] = None
    image: Optional[DecorationImage] = None
    border: Optional[Border] = None
    border_radius: Optional[BorderRadiusValue] = None
    shadow: Union[None, BoxShadow, List[BoxShadow]] = None
    gradient: Optional[Gradient] = None
    shape: Optional[BoxShape] = None
    blend_mode: Optional[BlendMode] = None


@dataclass
class BoxConstraints:
    min_width: Number = 0
    min_height: Number = 0
    max_width: Number = float("inf")
    max_height: Number = float("inf")

    def __post_init__(self):
        assert (
            0 <= self.min_width <= self.max_width <= float("inf")
        ), "min_width and max_width must be between 0 and infinity and min_width must be less than or equal to max_width"
        assert (
            0 <= self.min_height <= self.max_height <= float("inf")
        ), "min_height and max_height must be between 0 and infinity and min_height must be less than or equal to max_height"



from enum import Enum
from typing import Any, Dict, List, Optional

from flet.core.control import OptionalNumber
from flet.core.map.map import MapLatitudeLongitudeBounds
from flet.core.map.map_layer import MapLayer
from flet.core.ref import Ref
from flet.core.types import OptionalControlEventCallable


class MapTileLayerEvictErrorTileStrategy(Enum):
    DISPOSE = "dispose"
    NOT_VISIBLE = "notVisible"
    NOT_VISIBLE_RESPECT_MARGIN = "notVisibleRespectMargin"


class TileLayer(MapLayer):
    """
    The Map's main layer.
    Displays square raster images in a continuous grid, sourced from the provided utl_template.

    -----

    Online docs: https://flet.dev/docs/controls/maptilelayer
    """

    def __init__(
        self,
        url_template: str,
        fallback_url: Optional[str] = None,
        subdomains: Optional[List[str]] = None,
        tile_bounds: Optional[MapLatitudeLongitudeBounds] = None,
        tile_size: OptionalNumber = None,
        min_native_zoom: Optional[int] = None,
        max_native_zoom: Optional[int] = None,
        zoom_reverse: Optional[bool] = None,
        zoom_offset: OptionalNumber = None,
        keep_buffer: Optional[int] = None,
        pan_buffer: Optional[int] = None,
        enable_tms: Optional[bool] = None,
        keep_alive: Optional[bool] = None,
        enable_retina_mode: Optional[bool] = None,
        additional_options: Optional[Dict[str, str]] = None,
        max_zoom: OptionalNumber = None,
        min_zoom: OptionalNumber = None,
        error_image_src: Optional[str] = None,
        evict_error_tile_strategy: Optional[MapTileLayerEvictErrorTileStrategy] = None,
        on_image_error: OptionalControlEventCallable = None,
        #
        # MapLayer
        #
        ref: Optional[Ref] = None,
        visible: Optional[bool] = None,
        data: Any = None,
    ):

        MapLayer.__init__(
            self,
            ref=ref,
            visible=visible,
            data=data,
        )

        self.url_template = url_template
        self.fallback_url = fallback_url
        self.tile_size = tile_size
        self.min_native_zoom = min_native_zoom
        self.max_native_zoom = max_native_zoom
        self.zoom_reverse = zoom_reverse
        self.zoom_offset = zoom_offset
        self.keep_buffer = keep_buffer
        self.pan_buffer = pan_buffer
        self.enable_tms = enable_tms
        self.keep_alive = keep_alive
        self.max_zoom = max_zoom
        self.min_zoom = min_zoom
        self.error_image_src = error_image_src
        self.enable_retina_mode = enable_retina_mode
        self.on_image_error = on_image_error
        self.tile_bounds = tile_bounds
        self.evict_error_tile_strategy = evict_error_tile_strategy
        self.subdomains = subdomains
        self.additional_options = additional_options

    def _get_control_name(self):
        return "map_tile_layer"

    def before_update(self):
        super().before_update()
        assert self.url_template, "url_template is required"
        if isinstance(self.__tile_bounds, MapLatitudeLongitudeBounds):
            self._set_attr_json("tileBounds", self.__tile_bounds)
        if isinstance(self.__subdomains, list):
            self._set_attr_json("subdomains", self.__subdomains)
        if isinstance(self.__additional_options, dict):
            self._set_attr_json("additionalOptions", self.__additional_options)

    # url_template
    @property
    def url_template(self) -> str:
        return self._get_attr("urlTemplate")

    @url_template.setter
    def url_template(self, value: str):
        self._set_attr("urlTemplate", value)

    # fallback_url
    @property
    def fallback_url(self) -> Optional[str]:
        return self._get_attr("fallbackUrl")

    @fallback_url.setter
    def fallback_url(self, value: Optional[str]):
        self._set_attr("fallbackUrl", value)

    # subdomains
    @property
    def subdomains(self) -> Optional[List[str]]:
        return self.__subdomains

    @subdomains.setter
    def subdomains(self, value: Optional[List[str]]):
        self.__subdomains = value

    # additional_options
    @property
    def additional_options(self) -> Optional[Dict[str, str]]:
        return self.__additional_options

    @additional_options.setter
    def additional_options(self, value: Optional[Dict[str, str]]):
        self.__additional_options = value

    # tile_bounds
    @property
    def tile_bounds(self) -> Optional[MapLatitudeLongitudeBounds]:
        return self.__tile_bounds

    @tile_bounds.setter
    def tile_bounds(self, value: Optional[MapLatitudeLongitudeBounds]):
        self.__tile_bounds = value

    # tile_size
    @property
    def tile_size(self) -> float:
        return self._get_attr("tileSize", data_type="float", def_value=256.0)

    @tile_size.setter
    def tile_size(self, value: OptionalNumber):
        assert value is None or value >= 0, "tile_size cannot be negative"
        self._set_attr("tileSize", value)

    # min_native_zoom
    @property
    def min_native_zoom(self) -> int:
        return self._get_attr("minNativeZoom", data_type="int", def_value=0.0)

    @min_native_zoom.setter
    def min_native_zoom(self, value: Optional[int]):
        assert value is None or value >= 0, "min_native_zoom cannot be negative"
        self._set_attr("minNativeZoom", value)

    # max_native_zoom
    @property
    def max_native_zoom(self) -> int:
        return self._get_attr("maxNativeZoom", data_type="int", def_value=19)

    @max_native_zoom.setter
    def max_native_zoom(self, value: Optional[int]):
        assert value is None or value >= 0, "max_native_zoom cannot be negative"
        self._set_attr("maxNativeZoom", value)

    # zoom_reverse
    @property
    def zoom_reverse(self) -> bool:
        return self._get_attr("zoomReverse", data_type="bool", def_value=False)

    @zoom_reverse.setter
    def zoom_reverse(self, value: Optional[bool]):
        self._set_attr("zoomReverse", value)

    # zoom_offset
    @property
    def zoom_offset(self) -> float:
        return self._get_attr("zoomOffset", data_type="float", def_value=0.0)

    @zoom_offset.setter
    def zoom_offset(self, value: OptionalNumber):
        assert value is None or value >= 0, "zoom_offset cannot be negative"
        self._set_attr("zoomOffset", value)

    # keep_buffer
    @property
    def keep_buffer(self) -> int:
        return self._get_attr("keepBuffer", data_type="int", def_value=2)

    @keep_buffer.setter
    def keep_buffer(self, value: Optional[int]):
        self._set_attr("keepBuffer", value)

    # pan_buffer
    @property
    def pan_buffer(self) -> int:
        return self._get_attr("panBuffer", data_type="int", def_value=2)

    @pan_buffer.setter
    def pan_buffer(self, value: Optional[int]):
        self._set_attr("panBuffer", value)

    # enable_tms
    @property
    def enable_tms(self) -> bool:
        return self._get_attr("enableTms", data_type="bool", def_value=False)

    @enable_tms.setter
    def enable_tms(self, value: Optional[bool]):
        self._set_attr("enableTms", value)

    # enable_retina_mode
    @property
    def enable_retina_mode(self) -> bool:
        return self._get_attr("enableRetinaMode", data_type="bool", def_value=False)

    @enable_retina_mode.setter
    def enable_retina_mode(self, value: Optional[bool]):
        self._set_attr("enableRetinaMode", value)

    # max_zoom
    @property
    def max_zoom(self) -> float:
        return self._get_attr("maxZoom", data_type="float", def_value=float("inf"))

    @max_zoom.setter
    def max_zoom(self, value: OptionalNumber):
        assert value is None or value >= 0, "max_zoom cannot be negative"
        self._set_attr("maxZoom", value)

    # min_zoom
    @property
    def min_zoom(self) -> float:
        return self._get_attr("minZoom", data_type="float", def_value=0.0)

    @min_zoom.setter
    def min_zoom(self, value: OptionalNumber):
        assert value is None or value >= 0, "min_zoom cannot be negative"
        self._set_attr("minZoom", value)

    # error_image_src
    @property
    def error_image_src(self) -> Optional[str]:
        return self._get_attr("errorImageSrc")

    @error_image_src.setter
    def error_image_src(self, value: Optional[str]):
        self._set_attr("errorImageSrc", value)

    # evict_error_tile_strategy
    @property
    def evict_error_tile_strategy(self) -> Optional[MapTileLayerEvictErrorTileStrategy]:
        return self.__evict_error_tile_strategy

    @evict_error_tile_strategy.setter
    def evict_error_tile_strategy(
        self, value: Optional[MapTileLayerEvictErrorTileStrategy]
    ):
        self.__evict_error_tile_strategy = value
        self._set_enum_attr(
            "evictErrorTileStrategy", value, MapTileLayerEvictErrorTileStrategy
        )

    # on_image_error
    @property
    def on_image_error(self) -> OptionalControlEventCallable:
        return self._get_event_handler("imageError")

    @on_image_error.setter
    def on_image_error(self, handler: OptionalControlEventCallable):
        self._add_event_handler("imageError", handler)



from flet.core.canvas.arc import Arc
from flet.core.canvas.canvas import Canvas, CanvasResizeEvent
from flet.core.canvas.circle import Circle
from flet.core.canvas.color import Color
from flet.core.canvas.fill import Fill
from flet.core.canvas.line import Line
from flet.core.canvas.oval import Oval
from flet.core.canvas.path import Path
from flet.core.canvas.points import PointMode, Points
from flet.core.canvas.rect import Rect
from flet.core.canvas.shadow import Shadow
from flet.core.canvas.text import Text



from typing import Any, Optional

from flet.core.canvas.shape import Shape
from flet.core.types import BlendMode, ColorEnums, ColorValue


class Color(Shape):
    def __init__(
        self,
        color: Optional[ColorValue] = None,
        blend_mode: Optional[BlendMode] = None,
        #
        # Control
        #
        ref=None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        Shape.__init__(self, ref=ref, visible=visible, disabled=disabled, data=data)

        self.color = color
        self.blend_mode = blend_mode

    def _get_control_name(self):
        return "color"

    def before_update(self):
        super().before_update()

    # color
    @property
    def color(self) -> Optional[ColorValue]:
        return self.__color

    @color.setter
    def color(self, value: Optional[ColorValue]):
        self.__color = value
        self._set_enum_attr("color", value, ColorEnums)

    # blend_mode
    @property
    def blend_mode(self) -> Optional[BlendMode]:
        return self.__blend_mode

    @blend_mode.setter
    def blend_mode(self, value: Optional[BlendMode]):
        self.__blend_mode = value
        self._set_attr(
            "blendMode", value.value if isinstance(value, BlendMode) else value
        )



from flet.auth.oauth_provider import OAuthProvider


class Auth0OAuthProvider(OAuthProvider):
    def __init__(
        self, domain: str, client_id: str, client_secret: str, redirect_url: str
    ) -> None:
        super().__init__(
            client_id=client_id,
            client_secret=client_secret,
            authorization_endpoint=f"https://{domain}/authorize",
            token_endpoint=f"https://{domain}/oauth/token",
            redirect_url=redirect_url,
            scopes=["offline_access"],
            user_scopes=["openid", "profile", "email"],
            user_endpoint=f"https://{domain}/userinfo",
            user_id_fn=lambda u: u["sub"],
            group_scopes=[],
        )
        self.domain = domain



from dataclasses import dataclass, field
from typing import List, Optional

from flet.core.types import ColorValue


@dataclass
class ChartPointLine:
    color: Optional[ColorValue] = field(default=None)
    width: Optional[float] = field(default=None)
    dash_pattern: Optional[List[int]] = field(default=None)



import json
from typing import Any, Callable, Optional, Union

from flet.core.adaptive_control import AdaptiveControl
from flet.core.alignment import Alignment
from flet.core.animation import AnimationValue
from flet.core.badge import BadgeValue
from flet.core.constrained_control import ConstrainedControl
from flet.core.control import Control, OptionalNumber
from flet.core.control_event import ControlEvent
from flet.core.event_handler import EventHandler
from flet.core.ref import Ref
from flet.core.tooltip import TooltipValue
from flet.core.types import (
    ClipBehavior,
    DurationValue,
    MarginValue,
    Number,
    Offset,
    OffsetValue,
    OptionalControlEventCallable,
    OptionalEventCallable,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
)


class InteractiveViewerInteractionStartEvent(ControlEvent):
    def __init__(self, e: ControlEvent):
        super().__init__(e.target, e.name, e.data, e.control, e.page)
        d = json.loads(e.data)
        self.pointer_count: int = d.get("pc")
        self.global_focal_point: Offset = Offset(d.get("fp_x"), d.get("fp_y"))
        self.local_focal_point: Offset = Offset(d.get("lfp_x"), d.get("lfp_y"))


class InteractiveViewerInteractionUpdateEvent(ControlEvent):
    def __init__(self, e: ControlEvent):
        super().__init__(e.target, e.name, e.data, e.control, e.page)
        d = json.loads(e.data)
        self.pointer_count: int = d.get("pc")
        self.global_focal_point: Offset = Offset(d.get("fp_x"), d.get("fp_y"))
        self.local_focal_point: Offset = Offset(d.get("lfp_x"), d.get("lfp_y"))
        self.scale: float = d.get("s")
        self.horizontal_scale: float = d.get("hs")
        self.vertical_scale: float = d.get("vs")
        self.rotation: float = d.get("rot")


class InteractiveViewerInteractionEndEvent(ControlEvent):
    def __init__(self, e: ControlEvent):
        super().__init__(e.target, e.name, e.data, e.control, e.page)
        d = json.loads(e.data)
        self.pointer_count: int = d.get("pc")
        self.scale_velocity: float = d.get("sv")


class InteractiveViewer(ConstrainedControl, AdaptiveControl):
    """
    InteractiveViewer allows users to pan, zoom, and rotate content.

    -----

    Online docs: https://flet.dev/docs/controls/interactiveviewer
    """

    def __init__(
        self,
        content: Control,
        pan_enabled: Optional[bool] = None,
        scale_enabled: Optional[bool] = None,
        trackpad_scroll_causes_scale: Optional[bool] = None,
        constrained: Optional[bool] = None,
        max_scale: OptionalNumber = None,
        min_scale: OptionalNumber = None,
        interaction_end_friction_coefficient: OptionalNumber = None,
        scale_factor: OptionalNumber = None,
        clip_behavior: Optional[ClipBehavior] = None,
        alignment: Optional[Alignment] = None,
        boundary_margin: Optional[MarginValue] = None,
        interaction_update_interval: Optional[int] = None,
        on_interaction_start: Optional[
            Callable[[InteractiveViewerInteractionStartEvent], None]
        ] = None,
        on_interaction_update: Optional[
            Callable[[InteractiveViewerInteractionUpdateEvent], None]
        ] = None,
        on_interaction_end: Optional[
            Callable[[InteractiveViewerInteractionEndEvent], None]
        ] = None,
        #
        # ConstrainedControl and AdaptiveControl
        #
        ref: Optional[Ref] = None,
        key: Optional[str] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        left: OptionalNumber = None,
        top: OptionalNumber = None,
        right: OptionalNumber = None,
        bottom: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: OptionalControlEventCallable = None,
        tooltip: Optional[TooltipValue] = None,
        badge: Optional[BadgeValue] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
        adaptive: Optional[bool] = None,
    ):
        ConstrainedControl.__init__(
            self,
            ref=ref,
            key=key,
            width=width,
            height=height,
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            tooltip=tooltip,
            badge=badge,
            visible=visible,
            disabled=disabled,
            data=data,
        )

        AdaptiveControl.__init__(self, adaptive=adaptive)

        self.__on_interaction_start = EventHandler(
            lambda e: InteractiveViewerInteractionStartEvent(e)
        )
        self._add_event_handler(
            "interaction_start", self.__on_interaction_start.get_handler()
        )
        self.__on_interaction_update = EventHandler(
            lambda e: InteractiveViewerInteractionUpdateEvent(e)
        )
        self._add_event_handler(
            "interaction_update", self.__on_interaction_update.get_handler()
        )
        self.__on_interaction_end = EventHandler(
            lambda e: InteractiveViewerInteractionEndEvent(e)
        )
        self._add_event_handler(
            "interaction_end", self.__on_interaction_end.get_handler()
        )

        self.content = content
        self.pan_enabled = pan_enabled
        self.scale_enabled = scale_enabled
        self.trackpad_scroll_causes_scale = trackpad_scroll_causes_scale
        self.constrained = constrained
        self.max_scale = max_scale
        self.min_scale = min_scale
        self.interaction_end_friction_coefficient = interaction_end_friction_coefficient
        self.scale_factor = scale_factor
        self.clip_behavior = clip_behavior
        self.alignment = alignment
        self.boundary_margin = boundary_margin
        self.on_interaction_start = on_interaction_start
        self.on_interaction_end = on_interaction_end
        self.on_interaction_update = on_interaction_update
        self.interaction_update_interval = interaction_update_interval

    def _get_control_name(self):
        return "interactiveviewer"

    def before_update(self):
        super().before_update()
        assert self.__content.visible, "content must be visible"
        assert (
            self.max_scale >= self.min_scale
        ), "max_scale must be greather than or equal to min_scale"
        self._set_attr_json("alignment", self.__alignment)
        self._set_attr_json("boundaryMargin", self.__boundary_margin)

    def _get_children(self):
        return [self.__content]

    def reset(self, animation_duration: Optional[DurationValue] = None):
        self.invoke_method(
            "reset", arguments={"duration": self._convert_attr_json(animation_duration)}
        )

    def save_state(self):
        self.invoke_method("save_state")

    def restore_state(self):
        self.invoke_method("restore_state")

    def zoom(self, factor: Number):
        self.invoke_method("zoom", arguments={"factor": str(factor)})

    def pan(self, dx: Number, dy: Number):
        self.invoke_method("pan", arguments={"dx": str(dx), "dy": str(dy)})

    # min_scale
    @property
    def min_scale(self) -> float:
        return self._get_attr("minScale", data_type="float", def_value=0.8)

    @min_scale.setter
    def min_scale(self, value: OptionalNumber):
        assert value is None or value > 0, "min_scale must be greater than 0"
        self._set_attr("minScale", value)

    # interaction_update_interval
    @property
    def interaction_update_interval(self) -> int:
        return self._get_attr(
            "interactionUpdateInterval", data_type="int", def_value=200
        )

    @interaction_update_interval.setter
    def interaction_update_interval(self, value: Optional[int]):
        self._set_attr("interactionUpdateInterval", value)

    # max_scale
    @property
    def max_scale(self) -> float:
        return self._get_attr("maxScale", data_type="float", def_value=2.5)

    @max_scale.setter
    def max_scale(self, value: OptionalNumber):
        assert value is None or value > 0, "max_scale must be greater than 0"
        self._set_attr("maxScale", value)

    # interaction_end_friction_coefficient
    @property
    def interaction_end_friction_coefficient(self) -> float:
        return self._get_attr(
            "interactionEndFrictionCoefficient", data_type="float", def_value=0.0000135
        )

    @interaction_end_friction_coefficient.setter
    def interaction_end_friction_coefficient(self, value: OptionalNumber):
        assert (
            value is None or value > 0
        ), "interaction_end_friction_coefficient must be greater than 0"
        self._set_attr("interactionEndFrictionCoefficient", value)

    # content
    @property
    def content(self) -> Control:
        return self.__content

    @content.setter
    def content(self, value: Control):
        self.__content = value

    # pan_enabled
    @property
    def pan_enabled(self) -> bool:
        return self._get_attr("panEnabled", data_type="bool", def_value=True)

    @pan_enabled.setter
    def pan_enabled(self, value: Optional[bool]):
        self._set_attr("panEnabled", value)

    # scale_enabled
    @property
    def scale_enabled(self) -> bool:
        return self._get_attr("scaleEnabled", data_type="bool", def_value=True)

    @scale_enabled.setter
    def scale_enabled(self, value: Optional[bool]):
        self._set_attr("scaleEnabled", value)

    # trackpad_scroll_causes_scale
    @property
    def trackpad_scroll_causes_scale(self) -> bool:
        return self._get_attr(
            "trackpadScrollCausesScale", data_type="bool", def_value=False
        )

    @trackpad_scroll_causes_scale.setter
    def trackpad_scroll_causes_scale(self, value: Optional[bool]):
        self._set_attr("trackpadScrollCausesScale", value)

    # constrained
    @property
    def constrained(self) -> bool:
        return self._get_attr("constrained", data_type="bool", def_value=True)

    @constrained.setter
    def constrained(self, value: Optional[bool]):
        self._set_attr("constrained", value)

    # scale_factor
    @property
    def scale_factor(self) -> float:
        return self._get_attr("scaleFactor", data_type="float", def_value=200)

    @scale_factor.setter
    def scale_factor(self, value: OptionalNumber):
        self._set_attr("scaleFactor", value)

    # clip_behavior
    @property
    def clip_behavior(self) -> Optional[ClipBehavior]:
        return self.__clip_behavior

    @clip_behavior.setter
    def clip_behavior(self, value: Optional[ClipBehavior]):
        self.__clip_behavior = value
        self._set_enum_attr("clipBehavior", value, ClipBehavior)

    # alignment
    @property
    def alignment(self) -> Optional[Alignment]:
        return self.__alignment

    @alignment.setter
    def alignment(self, value: Optional[Alignment]):
        self.__alignment = value

    # boundary_margin
    @property
    def boundary_margin(self) -> Optional[MarginValue]:
        return self.__boundary_margin

    @boundary_margin.setter
    def boundary_margin(self, value: Optional[MarginValue]):
        self.__boundary_margin = value

    # on_interaction_start
    @property
    def on_interaction_start(
        self,
    ) -> OptionalEventCallable[InteractiveViewerInteractionStartEvent]:
        return self.__on_interaction_start.handler

    @on_interaction_start.setter
    def on_interaction_start(
        self,
        handler: OptionalEventCallable[InteractiveViewerInteractionStartEvent],
    ):
        self.__on_interaction_start.handler = handler

    # on_interaction_update
    @property
    def on_interaction_update(
        self,
    ) -> OptionalEventCallable[InteractiveViewerInteractionUpdateEvent]:
        return self.__on_interaction_update.handler

    @on_interaction_update.setter
    def on_interaction_update(
        self,
        handler: OptionalEventCallable[InteractiveViewerInteractionUpdateEvent],
    ):
        self.__on_interaction_update.handler = handler

    # on_interaction_end
    @property
    def on_interaction_end(
        self,
    ) -> OptionalEventCallable[InteractiveViewerInteractionEndEvent]:
        return self.__on_interaction_end.handler

    @on_interaction_end.setter
    def on_interaction_end(
        self, handler: OptionalEventCallable[InteractiveViewerInteractionEndEvent]
    ):
        self.__on_interaction_end.handler = handler



from typing import Any, Dict, List


class SessionStorage:
    def __init__(self, page):
        self.__page = page
        self.__store: Dict[str, Any] = {}

    def set(self, key: str, value: Any):
        self.__store[key] = value

    def get(self, key: str):
        return self.__store.get(key)

    def contains_key(self, key: str) -> bool:
        return key in self.__store

    def remove(self, key: str):
        self.__store.pop(key)

    def get_keys(self) -> List[str]:
        return list(self.__store.keys())

    def clear(self):
        self.__store.clear()



from enum import Enum
from typing import Any, Optional

from flet.core.control import Control
from flet.core.ref import Ref


class Assertiveness(Enum):
    POLITE = "polite"
    ASSERTIVE = "assertive"


class SemanticsService(Control):
    def __init__(
        self,
        ref: Optional[Ref] = None,
        data: Any = None,
    ):
        Control.__init__(
            self,
            ref=ref,
            data=data,
        )

    def _get_control_name(self):
        return "semanticsservice"

    def announce_message(
        self,
        message: str,
        rtl: bool = False,
        assertiveness: Assertiveness = Assertiveness.POLITE,
    ):
        self.invoke_method(
            "announce_message",
            arguments={
                "message": message,
                "rtl": str(rtl),
                "assertiveness": assertiveness.value
                if isinstance(assertiveness, Assertiveness)
                else str(assertiveness),
            },
        )

    def announce_tooltip(self, message: str):
        self.invoke_method(
            "announce_tooltip",
            arguments={
                "message": message,
            },
        )



from typing import Any, Optional, Union

from flet.core.animation import AnimationValue
from flet.core.badge import BadgeValue
from flet.core.constrained_control import ConstrainedControl
from flet.core.control import OptionalNumber
from flet.core.ref import Ref
from flet.core.tooltip import TooltipValue
from flet.core.types import (
    ColorEnums,
    ColorValue,
    ControlStateValue,
    IconValue,
    LabelPosition,
    OffsetValue,
    OptionalControlEventCallable,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
)
from flet.utils.deprecated import deprecated_property


class CupertinoSwitch(ConstrainedControl):
    """
    An iOS-style switch. Used to toggle the on/off state of a single setting.

    Example:
    ```
    import flet as ft

    def main(page: ft.Page):
        page.add(
            ft.CupertinoSwitch(label="Cupertino Switch", value=True),
            ft.Switch(label="Material Checkbox", value=True),
            ft.Container(height=20),
            ft.Text(
                "Adaptive Switch shows as CupertinoSwitch on macOS and iOS and as Switch on other platforms:"
            ),
            ft.Switch(adaptive=True, label="Adaptive Switch", value=True),
        )

    ft.app(target=main)
    ```
    -----

    Online docs: https://flet.dev/docs/controls/cupertinoswitch
    """

    def __init__(
        self,
        label: Optional[str] = None,
        value: Optional[bool] = None,
        label_position: Optional[LabelPosition] = None,
        active_color: Optional[ColorValue] = None,
        thumb_color: Optional[ColorValue] = None,
        track_color: Optional[ColorValue] = None,
        focus_color: Optional[ColorValue] = None,
        autofocus: Optional[bool] = None,
        on_label_color: Optional[ColorValue] = None,
        off_label_color: Optional[ColorValue] = None,
        active_thumb_image: Optional[str] = None,
        inactive_thumb_image: Optional[str] = None,
        active_track_color: Optional[ColorValue] = None,
        inactive_thumb_color: Optional[ColorValue] = None,
        inactive_track_color: Optional[ColorValue] = None,
        track_outline_color: ControlStateValue[ColorValue] = None,
        track_outline_width: ControlStateValue[OptionalNumber] = None,
        thumb_icon: ControlStateValue[IconValue] = None,
        on_change: OptionalControlEventCallable = None,
        on_focus: OptionalControlEventCallable = None,
        on_blur: OptionalControlEventCallable = None,
        on_image_error: OptionalControlEventCallable = None,
        #
        # ConstrainedControl
        #
        ref: Optional[Ref] = None,
        key: Optional[str] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        left: OptionalNumber = None,
        top: OptionalNumber = None,
        right: OptionalNumber = None,
        bottom: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: OptionalControlEventCallable = None,
        tooltip: Optional[TooltipValue] = None,
        badge: Optional[BadgeValue] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        ConstrainedControl.__init__(
            self,
            ref=ref,
            key=key,
            width=width,
            height=height,
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            tooltip=tooltip,
            badge=badge,
            visible=visible,
            disabled=disabled,
            data=data,
        )
        self.value = value
        self.label = label
        self.label_position = label_position
        self.autofocus = autofocus
        self.active_color = active_color
        self.focus_color = focus_color
        self.thumb_color = thumb_color
        self.track_color = track_color
        self.on_change = on_change
        self.on_focus = on_focus
        self.on_blur = on_blur
        self.on_label_color = on_label_color
        self.off_label_color = off_label_color
        self.active_thumb_image = active_thumb_image
        self.active_track_color = active_track_color
        self.inactive_thumb_color = inactive_thumb_color
        self.inactive_track_color = inactive_track_color
        self.track_outline_color = track_outline_color
        self.track_outline_width = track_outline_width
        self.thumb_icon = thumb_icon
        self.inactive_thumb_image = inactive_thumb_image
        self.on_image_error = on_image_error

    def _get_control_name(self):
        return "cupertinoswitch"

    def before_update(self):
        super().before_update()
        self._set_attr_json("thumbColor", self.__thumb_color)
        self._set_attr_json("trackColor", self.__track_color)
        self._set_attr_json(
            "trackOutlineColor", self.__track_outline_color, wrap_attr_dict=True
        )
        self._set_attr_json(
            "trackOutlineWidth", self.__track_outline_width, wrap_attr_dict=True
        )
        self._set_attr_json("thumbIcon", self.__thumb_icon, wrap_attr_dict=True)

    # value
    @property
    def value(self) -> bool:
        return self._get_attr("value", data_type="bool", def_value=False)

    @value.setter
    def value(self, value: Optional[bool]):
        self._set_attr("value", value)

    # label
    @property
    def label(self) -> Optional[str]:
        return self._get_attr("label")

    @label.setter
    def label(self, value: Optional[str]):
        self._set_attr("label", value)

    # active_thumb_image
    @property
    def active_thumb_image(self) -> Optional[str]:
        return self._get_attr("activeThumbImage")

    @active_thumb_image.setter
    def active_thumb_image(self, value: Optional[str]):
        self._set_attr("activeThumbImage", value)

    # inactive_thumb_image
    @property
    def inactive_thumb_image(self) -> Optional[str]:
        return self._get_attr("inactiveThumbImage")

    @inactive_thumb_image.setter
    def inactive_thumb_image(self, value: Optional[str]):
        self._set_attr("inactiveThumbImage", value)

    # active_track_color
    @property
    def active_track_color(self) -> Optional[ColorValue]:
        return self.__active_track_color

    @active_track_color.setter
    def active_track_color(self, value: Optional[ColorValue]):
        self.__active_track_color = value
        self._set_enum_attr("activeTrackColor", value, ColorEnums)

    # inactive_track_color
    @property
    def inactive_track_color(self) -> Optional[ColorValue]:
        return self.__inactive_track_color

    @inactive_track_color.setter
    def inactive_track_color(self, value: Optional[ColorValue]):
        self.__inactive_track_color = value
        self._set_enum_attr("inactiveTrackColor", value, ColorEnums)

    # track_outline_color
    @property
    def track_outline_color(self) -> ControlStateValue[ColorValue]:
        return self.__track_outline_color

    @track_outline_color.setter
    def track_outline_color(self, value: ControlStateValue[ColorValue]):
        self.__track_outline_color = value

    # track_outline_width
    @property
    def track_outline_width(self) -> ControlStateValue[OptionalNumber]:
        return self.__track_outline_width

    @track_outline_width.setter
    def track_outline_width(self, value: ControlStateValue[OptionalNumber]):
        self.__track_outline_width = value

    # thumb_icon
    @property
    def thumb_icon(self) -> ControlStateValue[IconValue]:
        return self.__thumb_icon

    @thumb_icon.setter
    def thumb_icon(self, value: ControlStateValue[IconValue]):
        self.__thumb_icon = value

    # label_position
    @property
    def label_position(self) -> Optional[LabelPosition]:
        return self.__label_position

    @label_position.setter
    def label_position(self, value: Optional[LabelPosition]):
        self.__label_position = value
        self._set_enum_attr("labelPosition", value, LabelPosition)

    # autofocus
    @property
    def autofocus(self) -> bool:
        return self._get_attr("autofocus", data_type="bool", def_value=False)

    @autofocus.setter
    def autofocus(self, value: Optional[bool]):
        self._set_attr("autofocus", value)

    # active_color
    @property
    def active_color(self) -> Optional[ColorValue]:
        deprecated_property(
            name="active_color",
            reason="Use active_track_color instead.",
            version="0.26.0",
            delete_version="0.29.0",
        )
        return self.__active_color

    @active_color.setter
    def active_color(self, value: Optional[ColorValue]):
        self.__active_color = value
        self._set_enum_attr("activeColor", value, ColorEnums)
        if value is not None:
            deprecated_property(
                name="active_color",
                reason="Use active_track_color instead.",
                version="0.26.0",
                delete_version="0.29.0",
            )

    # focus_color
    @property
    def focus_color(self) -> Optional[ColorValue]:
        return self.__focus_color

    @focus_color.setter
    def focus_color(self, value: Optional[ColorValue]):
        self.__focus_color = value
        self._set_enum_attr("focusColor", value, ColorEnums)

    # thumb_color
    @property
    def thumb_color(self) -> Optional[ColorValue]:
        return self.__thumb_color

    @thumb_color.setter
    def thumb_color(self, value: Optional[ColorValue]):
        self.__thumb_color = value

    # track_color
    @property
    def track_color(self) -> Optional[ColorValue]:
        deprecated_property(
            name="track_color",
            reason="Use inactive_track_color instead.",
            version="0.26.0",
            delete_version="0.29.0",
        )
        return self.__track_color

    @track_color.setter
    def track_color(self, value: Optional[ColorValue]):
        self.__track_color = value
        if value is not None:
            deprecated_property(
                name="track_color",
                reason="Use inactive_track_color instead.",
                version="0.26.0",
                delete_version="0.29.0",
            )

    # on_label_color
    @property
    def on_label_color(self) -> Optional[ColorValue]:
        return self.__on_label_color

    @on_label_color.setter
    def on_label_color(self, value: Optional[ColorValue]):
        self.__on_label_color = value
        self._set_enum_attr("onLabelColor", value, ColorEnums)

    # off_label_color
    @property
    def off_label_color(self) -> Optional[ColorValue]:
        return self.__off_label_color

    @off_label_color.setter
    def off_label_color(self, value: Optional[ColorValue]):
        self.__off_label_color = value
        self._set_enum_attr("offLabelColor", value, ColorEnums)

    # on_change
    @property
    def on_change(self) -> OptionalControlEventCallable:
        return self._get_event_handler("change")

    @on_change.setter
    def on_change(self, handler: OptionalControlEventCallable):
        self._add_event_handler("change", handler)

    # on_focus
    @property
    def on_focus(self) -> OptionalControlEventCallable:
        return self._get_event_handler("focus")

    @on_focus.setter
    def on_focus(self, handler: OptionalControlEventCallable):
        self._add_event_handler("focus", handler)

    # on_blur
    @property
    def on_blur(self) -> OptionalControlEventCallable:
        return self._get_event_handler("blur")

    @on_blur.setter
    def on_blur(self, handler: OptionalControlEventCallable):
        self._add_event_handler("blur", handler)

    # on_image_error
    @property
    def on_image_error(self) -> OptionalControlEventCallable:
        return self._get_event_handler("image_error")

    @on_image_error.setter
    def on_image_error(self, handler: OptionalControlEventCallable):
        self._add_event_handler("image_error", handler)



from typing import Any, Optional, Union

from flet.core.animation import AnimationValue
from flet.core.badge import BadgeValue
from flet.core.box import BoxConstraints
from flet.core.constrained_control import ConstrainedControl
from flet.core.control import OptionalNumber
from flet.core.ref import Ref
from flet.core.tooltip import TooltipValue
from flet.core.types import (
    ColorEnums,
    ColorValue,
    OffsetValue,
    OptionalControlEventCallable,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
    StrokeCap,
    PaddingValue,
)
from flet.utils.deprecated import deprecated_property


class ProgressRing(ConstrainedControl):
    """
    A material design circular progress indicator, which spins to indicate that the application is busy.

    A control that shows progress along a circle.

    Example:

    ```
    from time import sleep
    import flet as ft

    def main(page: ft.Page):
        pr = ft.ProgressRing(width=16, height=16, stroke_width = 2)

        page.add(
            ft.Text("Circular progress indicator", style="headlineSmall"),
            ft.Row([pr, ft.Text("Wait for the completion...")]),
            ft.Text("Indeterminate circular progress", style="headlineSmall"),
            ft.Column(
                [ft.ProgressRing(), ft.Text("I'm going to run for ages...")],
                horizontal_alignment=ft.CrossAxisAlignment.CENTER,
            ),
        )

        for i in range(0, 101):
            pr.value = i * 0.01
            sleep(0.1)
            page.update()

    ft.app(target=main)
    ```

    -----

    Online docs: https://flet.dev/docs/controls/progressring
    """

    def __init__(
        self,
        value: OptionalNumber = None,
        stroke_width: OptionalNumber = None,
        color: Optional[ColorValue] = None,
        bgcolor: Optional[ColorValue] = None,
        stroke_align: OptionalNumber = None,
        stroke_cap: Optional[StrokeCap] = None,
        semantics_label: Optional[str] = None,
        semantics_value: OptionalNumber = None,
        track_gap: OptionalNumber = None,
        size_constraints: Optional[BoxConstraints] = None,
        padding: Optional[PaddingValue] = None,
        year_2023: Optional[bool] = None,
        #
        # ConstrainedControl
        #
        ref: Optional[Ref] = None,
        key: Optional[str] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        left: OptionalNumber = None,
        top: OptionalNumber = None,
        right: OptionalNumber = None,
        bottom: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: OptionalControlEventCallable = None,
        tooltip: Optional[TooltipValue] = None,
        badge: Optional[BadgeValue] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        ConstrainedControl.__init__(
            self,
            ref=ref,
            key=key,
            width=width,
            height=height,
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            tooltip=tooltip,
            badge=badge,
            visible=visible,
            disabled=disabled,
            data=data,
        )
        self.value = value
        self.stroke_width = stroke_width
        self.color = color
        self.bgcolor = bgcolor
        self.semantics_label = semantics_label
        self.semantics_value = semantics_value
        self.stroke_align = stroke_align
        self.stroke_cap = stroke_cap
        self.track_gap = track_gap
        self.size_constraints = size_constraints
        self.padding = padding
        self.year_2023 = year_2023

    def _get_control_name(self):
        return "progressring"

    def before_update(self):
        super().before_update()
        self._set_attr_json("padding", self.__padding)
        self._set_attr_json("sizeConstraints", self.__size_constraints)

    # value
    @property
    def value(self) -> OptionalNumber:
        return self._get_attr("value", data_type="float")

    @value.setter
    def value(self, value: OptionalNumber):
        self._set_attr("value", value)

    # semantics_label
    @property
    def semantics_label(self) -> Optional[str]:
        return self._get_attr("semanticsLabel")

    @semantics_label.setter
    def semantics_label(self, value: Optional[str]):
        self._set_attr("semanticsLabel", value)

    # track_gap
    @property
    def track_gap(self) -> OptionalNumber:
        return self._get_attr("trackGap", data_type="float")

    @track_gap.setter
    def track_gap(self, value: OptionalNumber):
        self._set_attr("trackGap", value)

    # padding
    @property
    def padding(self) -> Optional[PaddingValue]:
        return self.__padding

    @padding.setter
    def padding(self, value: Optional[PaddingValue]):
        self.__padding = value

    # size_constraints
    @property
    def size_constraints(self) -> Optional[BoxConstraints]:
        return self.__size_constraints

    @size_constraints.setter
    def size_constraints(self, value: Optional[BoxConstraints]):
        self.__size_constraints = value

    # year_2023
    @property
    def year_2023(self) -> Optional[bool]:
        deprecated_property(
            name="year_2023",
            version="0.27.0",
            delete_version=None,  # not known for now
            reason="Set this flag to False to opt into the 2024 Slider appearance. In the future, this flag will default to False.",
        )
        return self._get_attr("year2023", data_type="bool", def_value=True)

    @year_2023.setter
    def year_2023(self, value: Optional[bool]):
        self._set_attr("year2023", value)
        if value is not None:
            deprecated_property(
                name="year_2023",
                version="0.27.0",
                delete_version=None,  # not known for now
                reason="Set this flag to False to opt into the 2024 Slider appearance. In the future, this flag will default to False.",
            )

    # stroke_width
    @property
    def stroke_width(self) -> float:
        return self._get_attr("strokeWidth", data_type="float", def_value=4.0)

    @stroke_width.setter
    def stroke_width(self, value: OptionalNumber):
        assert value is None or value >= 0, "stroke_width cannot be negative"
        self._set_attr("strokeWidth", value)

    # stroke_align
    @property
    def stroke_align(self) -> float:
        return self._get_attr("strokeAlign", data_type="float", def_value=0.0)

    @stroke_align.setter
    def stroke_align(self, value: OptionalNumber):
        self._set_attr("strokeAlign", value)

    # stroke_cap
    @property
    def stroke_cap(self) -> Optional[StrokeCap]:
        return self.__stroke_cap

    @stroke_cap.setter
    def stroke_cap(self, value: Optional[StrokeCap]):
        self.__stroke_cap = value
        self._set_enum_attr("strokeCap", value, StrokeCap)

    # color
    @property
    def color(self) -> Optional[ColorValue]:
        return self.__color

    @color.setter
    def color(self, value: Optional[ColorValue]):
        self.__color = value
        self._set_enum_attr("color", value, ColorEnums)

    # bgcolor
    @property
    def bgcolor(self) -> Optional[ColorValue]:
        return self.__bgcolor

    @bgcolor.setter
    def bgcolor(self, value: Optional[ColorValue]):
        self.__bgcolor = value
        self._set_enum_attr("bgcolor", value, ColorEnums)



import functools
import warnings
from typing import Optional


def deprecated(reason: str, version: str, delete_version: str, is_method=True):
    """
    A decorator function that marks a function/method/property/event as deprecated.

    :param reason: The reason for deprecation.
    :param version: The version from which the function was deprecated.
    :param delete_version: The version in which the function will be removed from the API.
    :param is_method: if the deprecated item is a method (True) or property/function/event (False)
    """

    def decorator(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            warnings.warn(
                f"{func.__name__}{'()' if is_method else ''} is deprecated since version {version} "
                f"and will be removed in version {delete_version}. {reason}",
                category=DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return new_func

    return decorator


def deprecated_class(reason: str, version: str, delete_version: str):
    def decorator(cls):
        msg = f"{cls.__name__} is deprecated since version {version} and will be removed in version {delete_version}. {reason}"

        # Wrap the original __init__ method
        orig_init = cls.__init__

        @functools.wraps(orig_init)
        def new_init(self, *args, **kwargs):
            warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
            orig_init(self, *args, **kwargs)

        cls.__init__ = new_init
        return cls

    return decorator


def deprecated_property(
    name: str, reason: str, version: str, delete_version: Optional[str] = None
):
    warnings.warn(
        f"{name} property is deprecated since version {version}"
        f"{' and will be removed in version ' + delete_version if delete_version else ''}. {reason}",
        category=DeprecationWarning,
        stacklevel=2,
    )



from flet.core.elevated_button import ElevatedButton


class FilledTonalButton(ElevatedButton):
    """
    A filled tonal button is an alternative middle ground between FilledButton and OutlinedButton buttons. They’re useful in contexts where a lower-priority button requires slightly more emphasis than an outline would give, such as "Next" in an onboarding flow. Tonal buttons use the secondary color mapping.

    Example:
    ```
    import flet as ft


    def main(page: ft.Page):
        page.title = "Basic filled tonal buttons"
        page.add(
            ft.FilledTonalButton(text="Filled tonal button"),
            ft.FilledTonalButton("Disabled button", disabled=True),
            ft.FilledTonalButton("Button with icon", icon="add"),
        )

    ft.app(target=main)
    ```

    -----

    Online docs: https://flet.dev/docs/controls/filledtonalbutton
    """

    def _get_control_name(self):
        return "filledtonalbutton"



import warnings
from typing import Any, Optional, Union

from flet.core.alignment import Alignment
from flet.core.animation import AnimationValue
from flet.core.badge import BadgeValue
from flet.core.constrained_control import ConstrainedControl
from flet.core.control import Control, OptionalNumber
from flet.core.ref import Ref
from flet.core.tooltip import TooltipValue
from flet.core.types import (
    BorderRadiusValue,
    ColorEnums,
    ColorValue,
    IconEnums,
    IconValue,
    OffsetValue,
    OptionalControlEventCallable,
    PaddingValue,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
    UrlTarget,
)


class CupertinoButton(ConstrainedControl):
    """
    An iOS-style button.

    -----

    Online docs: https://flet.dev/docs/controls/cupertinobutton
    """

    def __init__(
        self,
        text: Optional[str] = None,
        icon: Optional[IconValue] = None,
        icon_color: Optional[ColorValue] = None,
        content: Optional[Control] = None,
        bgcolor: Optional[ColorValue] = None,
        color: Optional[ColorValue] = None,
        disabled_bgcolor: Optional[ColorValue] = None,
        opacity_on_click: OptionalNumber = None,
        min_size: OptionalNumber = None,
        padding: Optional[PaddingValue] = None,
        alignment: Optional[Alignment] = None,
        border_radius: Optional[BorderRadiusValue] = None,
        url: Optional[str] = None,
        url_target: Optional[UrlTarget] = None,
        autofocus: Optional[bool] = None,
        focus_color: Optional[bool] = None,
        on_click: OptionalControlEventCallable = None,
        on_long_press: OptionalControlEventCallable = None,
        on_focus: OptionalControlEventCallable = None,
        on_blur: OptionalControlEventCallable = None,
        #
        # ConstrainedControl
        #
        ref: Optional[Ref] = None,
        key: Optional[str] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        left: OptionalNumber = None,
        top: OptionalNumber = None,
        right: OptionalNumber = None,
        bottom: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: OptionalControlEventCallable = None,
        tooltip: Optional[TooltipValue] = None,
        badge: Optional[BadgeValue] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        ConstrainedControl.__init__(
            self,
            ref=ref,
            key=key,
            width=width,
            height=height,
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            tooltip=tooltip,
            badge=badge,
            visible=visible,
            disabled=disabled,
            data=data,
        )

        self.disabled_bgcolor = disabled_bgcolor
        self.text = text
        self.icon = icon
        self.icon_color = icon_color
        self.bgcolor = bgcolor
        self.color = color
        self.border_radius = border_radius
        self.min_size = min_size
        self.opacity_on_click = opacity_on_click
        self.padding = padding
        self.alignment = alignment
        self.content = content
        self.url = url
        self.url_target = url_target
        self.on_click = on_click
        self.on_long_press = on_long_press
        self.on_focus = on_focus
        self.on_blur = on_blur
        self.autofocus = autofocus
        self.focus_color = focus_color

    def _get_control_name(self):
        return "cupertinobutton"

    def before_update(self):
        super().before_update()
        assert (
            self.text or self.icon or (self.__content and self.__content.visible)
        ), "at minimum, text, icon or a visible content must be provided"
        self._set_attr_json("padding", self.__padding)
        self._set_attr_json("borderRadius", self.__border_radius)
        self._set_attr_json("alignment", self.__alignment)

    def _get_children(self):
        if self.__content is None:
            return []
        self.__content._set_attr_internal("n", "content")
        return [self.__content]

    # text
    @property
    def text(self):
        return self._get_attr("text")

    @text.setter
    def text(self, value):
        self._set_attr("text", value)

    # focus_color
    @property
    def focus_color(self):
        return self.__focus_color

    @focus_color.setter
    def focus_color(self, value):
        self.__focus_color = value
        self._set_enum_attr("focusColor", value, ColorEnums)

    # autofocus
    @property
    def autofocus(self):
        return self._get_attr("autofocus", data_type="bool", def_value=False)

    @autofocus.setter
    def autofocus(self, value):
        self._set_attr("autofocus", value)

    # icon
    @property
    def icon(self):
        return self.__icon

    @icon.setter
    def icon(self, value):
        self.__icon = value
        self._set_enum_attr("icon", value, IconEnums)

    # icon_color
    @property
    def icon_color(self):
        return self.__icon_color

    @icon_color.setter
    def icon_color(self, value):
        self.__icon_color = value
        self._set_enum_attr("iconColor", value, ColorEnums)

    # alignment
    @property
    def alignment(self) -> Optional[Alignment]:
        return self.__alignment

    @alignment.setter
    def alignment(self, value: Optional[Alignment]):
        self.__alignment = value

    # disabled_bgcolor
    @property
    def disabled_bgcolor(self) -> Optional[str]:
        return self.__disabled_bgcolor

    @disabled_bgcolor.setter
    def disabled_bgcolor(self, value: Optional[str]):
        self.__disabled_bgcolor = value
        self._set_enum_attr("disabledBgcolor", value, ColorEnums)

    # opacity_on_click
    @property
    def opacity_on_click(self) -> float:
        return self._get_attr("opacityOnClick", data_type="float", def_value=0.4)

    @opacity_on_click.setter
    def opacity_on_click(self, value: OptionalNumber):
        if value is not None:
            value = max(0.0, min(value, 1.0))  # make sure 0.0 <= value <= 1.0
        self._set_attr("opacityOnClick", value)

    # border_radius
    @property
    def border_radius(self) -> Optional[BorderRadiusValue]:
        return self.__border_radius

    @border_radius.setter
    def border_radius(self, value: Optional[BorderRadiusValue]):
        self.__border_radius = value

    # min_size
    @property
    def min_size(self) -> float:
        return self._get_attr("minSize", data_type="float", def_value=44.0)

    @min_size.setter
    def min_size(self, value: OptionalNumber):
        self._set_attr("minSize", value)

    # padding
    @property
    def padding(self) -> Optional[PaddingValue]:
        return self.__padding

    @padding.setter
    def padding(self, value: Optional[PaddingValue]):
        self.__padding = value

    # bgcolor
    @property
    def bgcolor(self) -> Optional[ColorValue]:
        return self.__bgcolor

    @bgcolor.setter
    def bgcolor(self, value: Optional[ColorValue]):
        self.__bgcolor = value
        self._set_enum_attr("bgColor", value, ColorEnums)

    # color
    @property
    def color(self) -> Optional[ColorValue]:
        return self.__color

    @color.setter
    def color(self, value: Optional[ColorValue]):
        self.__color = value
        self._set_enum_attr("color", value, ColorEnums)

    # url
    @property
    def url(self):
        return self._get_attr("url")

    @url.setter
    def url(self, value):
        self._set_attr("url", value)

    # url_target
    @property
    def url_target(self) -> Optional[UrlTarget]:
        return self.__url_target

    @url_target.setter
    def url_target(self, value: Optional[UrlTarget]):
        self.__url_target = value
        self._set_enum_attr("urlTarget", value, UrlTarget)

    # on_click
    @property
    def on_click(self) -> OptionalControlEventCallable:
        return self._get_event_handler("click")

    @on_click.setter
    def on_click(self, handler: OptionalControlEventCallable):
        self._add_event_handler("click", handler)

    # content
    @property
    def content(self) -> Optional[Control]:
        return self.__content

    @content.setter
    def content(self, value: Optional[Control]):
        self.__content = value

    # on_long_press
    @property
    def on_long_press(self) -> OptionalControlEventCallable:
        return self._get_event_handler("longPress")

    @on_long_press.setter
    def on_long_press(self, handler: OptionalControlEventCallable):
        self._add_event_handler("longPress", handler)

    # on_focus
    @property
    def on_focus(self) -> OptionalControlEventCallable:
        return self._get_event_handler("focus")

    @on_focus.setter
    def on_focus(self, handler: OptionalControlEventCallable):
        self._add_event_handler("focus", handler)

    # on_blur
    @property
    def on_blur(self) -> OptionalControlEventCallable:
        return self._get_event_handler("blur")

    @on_blur.setter
    def on_blur(self, handler: OptionalControlEventCallable):
        self._add_event_handler("blur", handler)



from typing import Any, Optional

from flet.core.adaptive_control import AdaptiveControl
from flet.core.control import Control
from flet.core.ref import Ref
from flet.core.types import IconEnums, IconValue, OptionalControlEventCallable


class CupertinoContextMenuAction(AdaptiveControl):
    """
    An action that can be added to a CupertinoContextMenu.

    -----

    Online docs: https://flet.dev/docs/controls/cupertinocontextmenuaction
    """

    def __init__(
        self,
        text: Optional[str] = None,
        content: Optional[Control] = None,
        is_default_action: Optional[bool] = None,
        is_destructive_action: Optional[bool] = None,
        trailing_icon: Optional[IconValue] = None,
        on_click: OptionalControlEventCallable = None,
        #
        # Control
        #
        ref: Optional[Ref] = None,
        disabled: Optional[bool] = None,
        visible: Optional[bool] = None,
        data: Any = None,
        adaptive: Optional[bool] = False,
    ):
        Control.__init__(
            self,
            ref=ref,
            disabled=disabled,
            visible=visible,
            data=data,
        )

        AdaptiveControl.__init__(self, adaptive=adaptive)

        self.is_default_action = is_default_action
        self.is_destructive_action = is_destructive_action
        self.content = content
        self.trailing_icon = trailing_icon
        self.on_click = on_click
        self.text = text

    def _get_control_name(self):
        return "cupertinocontextmenuaction"

    def _get_children(self):
        return [self.__content] if self.__content else []

    # is_default_action
    @property
    def is_default_action(self) -> bool:
        return self._get_attr("isDefaultAction", data_type="bool", def_value=False)

    @is_default_action.setter
    def is_default_action(self, value: Optional[bool]):
        self._set_attr("isDefaultAction", value)

    # is_destructive_action
    @property
    def is_destructive_action(self) -> bool:
        return self._get_attr("isDestructiveAction", data_type="bool", def_value=False)

    @is_destructive_action.setter
    def is_destructive_action(self, value: Optional[bool]):
        self._set_attr("isDestructiveAction", value)

    # trailing_icon
    @property
    def trailing_icon(self) -> Optional[IconValue]:
        return self.__trailing_icon

    @trailing_icon.setter
    def trailing_icon(self, value: Optional[IconValue]):
        self.__trailing_icon = value
        self._set_enum_attr("trailingIcon", value, IconEnums)

    # text
    @property
    def text(self) -> Optional[str]:
        return self._get_attr("text")

    @text.setter
    def text(self, value: Optional[str]):
        self._set_attr("text", value)

    # content
    @property
    def content(self) -> Optional[Control]:
        return self.__content

    @content.setter
    def content(self, value: Optional[Control]):
        self.__content = value

    # on_click
    @property
    def on_click(self) -> OptionalControlEventCallable:
        return self._get_event_handler("click")

    @on_click.setter
    def on_click(self, handler: OptionalControlEventCallable):
        self._add_event_handler("click", handler)



from typing import Any, Optional

from flet.core.canvas.shape import Shape
from flet.core.control import OptionalNumber
from flet.core.painting import Paint


class Circle(Shape):
    def __init__(
        self,
        x: OptionalNumber = None,
        y: OptionalNumber = None,
        radius: OptionalNumber = None,
        paint: Optional[Paint] = None,
        #
        # Control
        #
        ref=None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        Shape.__init__(self, ref=ref, visible=visible, disabled=disabled, data=data)

        self.x = x
        self.y = y
        self.radius = radius
        self.paint = paint

    def _get_control_name(self):
        return "circle"

    def before_update(self):
        super().before_update()
        self._set_attr_json("paint", self.__paint)

    # x
    @property
    def x(self) -> OptionalNumber:
        return self._get_attr("x")

    @x.setter
    def x(self, value: OptionalNumber):
        self._set_attr("x", value)

    # y
    @property
    def y(self) -> OptionalNumber:
        return self._get_attr("y")

    @y.setter
    def y(self, value: OptionalNumber):
        self._set_attr("y", value)

    # radius
    @property
    def radius(self) -> OptionalNumber:
        return self._get_attr("radius")

    @radius.setter
    def radius(self, value: OptionalNumber):
        self._set_attr("radius", value)

    # paint
    @property
    def paint(self) -> Optional[Paint]:
        return self.__paint

    @paint.setter
    def paint(self, value: Optional[Paint]):
        self.__paint = value
